import json
import unittest
from base64 import b64encode
from datetime import datetime, timezone
from unittest.mock import ANY

from django.test import ignore_warnings, override_settings, tag

from anymail.exceptions import AnymailImproperlyInstalled, AnymailNotSupportedWarning
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.sendgrid import SendGridTrackingWebhookView

from .utils_sendgrid import (
    ClientWithSendGridSignature,
    derive_public_webhook_key,
    make_key,
    sign,
)
from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase

CRYPTOGRAPHY_INSTALLED = ClientWithSendGridSignature is not None


# These tests run both with and without cryptograpy installed
@tag("sendgrid")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridWebhookSecurityTestCase(WebhookBasicAuthTestCase):
    def call_webhook(self):
        return self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps([]),
        )

    # Actual tests are in WebhookBasicAuthTestCase


@ignore_warnings(category=AnymailNotSupportedWarning)
class BaseSendGridWebhookTestCase(WebhookTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.raw_events = [
            {
                "email": "recipient@example.com",
                "timestamp": 1461095246,
                "anymail_id": "3c2f4df8-c6dd-4cd2-9b91-6582b81a0349",
                "smtp-id": "<wrfRRvF7Q0GgwUo2CvDmEA@ismtpd0006p1sjc2.sendgrid.net>",
                "sg_event_id": "ZyjAM5rnQmuI1KFInHQ3Nw",
                "sg_message_id": "wrfRRvF7Q0GgwUo2CvDmEA"
                ".filter0425p1mdw1.13037.57168B4A1D.0",
                "event": "processed",
                "category": ["tag1", "tag2"],
                "custom1": "value1",
                "custom2": "value2",
            }
        ]
        cls.data = json.dumps(cls.raw_events)


@tag("sendgrid")
@unittest.skipIf(
    CRYPTOGRAPHY_INSTALLED,
    "Uninstall 'cryptography' to test graceful error handling in SendGrid webhook tests",
)
class SendGridSignedWebhookMissingCryptographyTestCase(BaseSendGridWebhookTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.creds = [("cred1", "pass1"), ("cred1", "pass2")]

    @override_settings(
        ANYMAIL={
            "SENDGRID_TRACKING_WEBHOOK_VERIFICATION_KEY": (
                "MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE0ObrPrfwL/37SbZ52HNp4Gq0j2"
                "j4Nu8LvcNuazePhgFJlNnuk20Efm67BHVVQVNUGyPDrchn8IvF24ccppLpbw=="
            )
        }
    )
    def test_configured_verification_key_without_cryptography(self):
        with self.assertRaisesMessage(AnymailImproperlyInstalled, "cryptography"):
            self.client.post(
                "/anymail/sendgrid/tracking/",
                content_type="application/json",
                data=self.raw_events,
                HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="message timestamp",
                HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=b64encode(
                    "invalid".encode("utf-8")
                ),
            )

    @override_settings(ANYMAIL={"WEBHOOK_SECRET": ["cred1:pass1", "cred2:pass2"]})
    def test_unconfigured_verification_key_but_with_webhook_secret(self):
        for creds in self.creds:
            with self.subTest(creds=creds):
                self.set_basic_auth(*creds)
                self.client.post(
                    "/anymail/sendgrid/tracking/",
                    content_type="application/json",
                    data=self.raw_events,
                    HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="message timestamp",
                    HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=b64encode(
                        "invalid".encode("utf-8")
                    ),
                )


class BaseSendGridSignedWebhookSecurityTestCase(BaseSendGridWebhookTestCase):
    client_class = ClientWithSendGridSignature

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.private_key = make_key()

    def setUp(self):
        super().setUp()
        self.client.set_private_key(self.private_key)


@tag("sendgrid")
@unittest.skipUnless(
    CRYPTOGRAPHY_INSTALLED,
    "Install 'cryptography' to test SendGrid signature verification",
)
class SendGridSignedWebhookSecurityTestCase(BaseSendGridSignedWebhookSecurityTestCase):
    def setUp(self):
        super().setUp()
        self.clear_basic_auth()

    def test_invalid_signature(self):
        # This also verifies that the error log references the correct setting to check.
        with self.assertLogs() as logs:
            response = self.client.post(
                "/anymail/sendgrid/tracking/",
                content_type="application/json",
                data=self.raw_events,
                HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="message timestamp",
                HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=b64encode(
                    "invalid".encode("utf-8")
                ),
            )
        self.assertEqual(response.status_code, 400)
        self.assertIn(
            "check Anymail SENDGRID_TRACKING_WEBHOOK_VERIFICATION_KEY", logs.output[0]
        )

    def test_invalid_timestamp(self):
        timestamp = "data timestamp"
        signature = sign(self.private_key, timestamp, message=self.data)

        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="invalid data timestamp",
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=b64encode(signature),
        )
        self.assertEqual(response.status_code, 400)

    def test_invalid_timestamp_and_invalid_signature(self):
        timestamp = "data timestamp"
        different_private_key = make_key()
        incorrect_signature = sign(different_private_key, timestamp, message=self.data)

        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="invalid data timestamp",
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=b64encode(incorrect_signature),
        )
        self.assertEqual(response.status_code, 400)

    def test_empty_signature_(self):
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="message timestamp",
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE="",
        )
        self.assertEqual(response.status_code, 400)

    def test_empty_timestamp(self):
        timestamp = "data timestamp"
        signature = sign(self.private_key, timestamp, message=self.data)

        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="",
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=b64encode(signature),
        )
        self.assertEqual(response.status_code, 400)

    def test_empty_timestamp_and_empty_signature(self):
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="",
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE="",
        )
        self.assertEqual(response.status_code, 400)

    def test_missing_signature(self):
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="message timestamp",
        )
        self.assertEqual(response.status_code, 400)

    def test_missing_timestamp(self):
        # Should be a valid signature
        signature = sign(self.private_key, "", message=self.data)

        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=b64encode(signature),
        )
        self.assertEqual(response.status_code, 400)

    def test_missing_timestamp_and_missing_signature(self):
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            automatically_set_timestamp_and_signature_headers=False,
        )
        self.assertEqual(response.status_code, 400)

    def test_successful_signature_check_without_http_auth(self):
        timestamp = "data timestamp"
        signature = b64encode(sign(self.private_key, timestamp, message=self.data))

        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=self.raw_events,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP=timestamp,
            HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=signature,
        )
        self.assertEqual(response.status_code, 200)

    def test_verification_key_view_params(self):
        """Webhook verification key can be provided as a view param"""
        key = make_key()
        view = SendGridTrackingWebhookView.as_view(
            tracking_webhook_verification_key=derive_public_webhook_key(key),
        )
        view_instance = view.view_class(**view.view_initkwargs)
        self.assertEqual(
            view_instance.webhook_verification_key.public_numbers(),
            key.public_key().public_numbers(),
        )


@tag("sendgrid")
@unittest.skipUnless(
    CRYPTOGRAPHY_INSTALLED,
    "Install 'cryptography' to test SendGrid signature verification",
)
@override_settings(ANYMAIL={"WEBHOOK_SECRET": ["cred1:pass1", "cred2:pass2"]})
class SendGridSignedWebhookSecurityAndHttpBasicAuthTestCase(
    BaseSendGridSignedWebhookSecurityTestCase
):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.creds = [("cred1", "pass1"), ("cred1", "pass2")]

    def test_invalid_timestamp_and_invalid_signature_with_http_auth(self):
        timestamp = "data timestamp"

        for creds in self.creds:
            with self.subTest(creds=creds):
                private_key = make_key()
                signature = sign(private_key, timestamp, message=self.data)

                self.set_basic_auth(*creds)

                response = self.client.post(
                    "/anymail/sendgrid/tracking/",
                    content_type="application/json",
                    data=self.raw_events,
                    HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP="invalid data timestamp",
                    HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=b64encode(signature),
                )
                self.assertEqual(response.status_code, 400)

    def test_missing_timestamp_and_missing_signature_with_http_auth(self):
        for creds in self.creds:
            with self.subTest(creds=creds):
                self.set_basic_auth(*creds)
                response = self.client.post(
                    "/anymail/sendgrid/tracking/",
                    content_type="application/json",
                    data=self.raw_events,
                    automatically_set_timestamp_and_signature_headers=False,
                )
                self.assertEqual(response.status_code, 400)

    def test_successful_signature_check_with_basic_http_auth(self):
        for creds in self.creds:
            with self.subTest(creds=creds):
                self.set_basic_auth(*creds)

                timestamp = "data timestamp"
                signature = b64encode(
                    sign(self.private_key, timestamp, message=self.data)
                )

                response = self.client.post(
                    "/anymail/sendgrid/tracking/",
                    content_type="application/json",
                    data=self.raw_events,
                    HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP=timestamp,
                    HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE=signature,
                )
                self.assertEqual(response.status_code, 200)


@tag("sendgrid")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridDeliveryTestCase(WebhookTestCase):
    def test_not_supported_warning(self):
        with self.assertWarns(
            AnymailNotSupportedWarning,
            msg="django-anymail has dropped official support for SendGrid.",
        ):
            self.client.post(
                "/anymail/sendgrid/tracking/",
                content_type="application/json",
                data=json.dumps([]),
            )

    def test_processed_event(self):
        raw_events = [
            {
                "email": "recipient@example.com",
                "timestamp": 1461095246,
                "anymail_id": "3c2f4df8-c6dd-4cd2-9b91-6582b81a0349",
                "smtp-id": "<wrfRRvF7Q0GgwUo2CvDmEA@ismtpd0006p1sjc2.sendgrid.net>",
                "sg_event_id": "ZyjAM5rnQmuI1KFInHQ3Nw",
                "sg_message_id": "wrfRRvF7Q0GgwUo2CvDmEA"
                ".filter0425p1mdw1.13037.57168B4A1D.0",
                "event": "processed",
                "category": ["tag1", "tag2"],
                "custom1": "value1",
                "custom2": "value2",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "queued")
        self.assertEqual(
            event.timestamp, datetime(2016, 4, 19, 19, 47, 26, tzinfo=timezone.utc)
        )
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "3c2f4df8-c6dd-4cd2-9b91-6582b81a0349")
        self.assertEqual(event.event_id, "ZyjAM5rnQmuI1KFInHQ3Nw")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.tags, ["tag1", "tag2"])
        self.assertEqual(event.metadata, {"custom1": "value1", "custom2": "value2"})

    def test_delivered_event(self):
        raw_events = [
            {
                "ip": "167.89.17.173",
                "response": "250 2.0.0 OK 1461095248 m143si2210036ioe.159 - gsmtp ",
                "smtp-id": "<wrfRRvF7Q0GgwUo2CvDmEA@ismtpd0006p1sjc2.sendgrid.net>",
                "sg_event_id": "nOSv8m0eTQ-vxvwNwt3fZQ",
                "sg_message_id": "wrfRRvF7Q0GgwUo2CvDmEA"
                ".filter0425p1mdw1.13037.57168B4A1D.0",
                "tls": 1,
                "event": "delivered",
                "email": "recipient@example.com",
                "timestamp": 1461095250,
                "anymail_id": "4ab185c2-0171-492f-9ce0-27de258efc99",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(
            event.timestamp, datetime(2016, 4, 19, 19, 47, 30, tzinfo=timezone.utc)
        )
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "4ab185c2-0171-492f-9ce0-27de258efc99")
        self.assertEqual(event.event_id, "nOSv8m0eTQ-vxvwNwt3fZQ")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.mta_response, "250 2.0.0 OK 1461095248 m143si2210036ioe.159 - gsmtp "
        )
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_dropped_invalid_event(self):
        raw_events = [
            {
                "email": "invalid@invalid",
                "anymail_id": "c74002d9-7ccb-4f67-8b8c-766cec03c9a6",
                "timestamp": 1461095250,
                "smtp-id": "<wrfRRvF7Q0GgwUo2CvDmEA@ismtpd0006p1sjc2.sendgrid.net>",
                "sg_event_id": "3NPOePGOTkeM_U3fgWApfg",
                "sg_message_id": "filter0093p1las1.9128.5717FB8127.0",
                "reason": "Invalid",
                "event": "dropped",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "c74002d9-7ccb-4f67-8b8c-766cec03c9a6")
        self.assertEqual(event.event_id, "3NPOePGOTkeM_U3fgWApfg")
        self.assertEqual(event.recipient, "invalid@invalid")
        self.assertEqual(event.reject_reason, "invalid")
        self.assertEqual(event.mta_response, None)

    def test_dropped_bounced(self):
        # https://www.twilio.com/docs/sendgrid/for-developers/tracking-events/event#dropped
        raw_events = [
            {
                "email": "example@example.com",
                "timestamp": 1513299569,
                "smtp-id": "<14c5d75ce93.dfd.64b469@ismtpd-555>",
                "event": "dropped",
                "category": "cat facts",
                "sg_event_id": "zmzJhfJgAfUSOW80yEbPyw==",
                "sg_message_id": "14c5d75ce93.dfd.64b469.filter0001.16648.5515E0B88.0",
                "reason": "Bounced Address",
                "status": "5.0.0",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.recipient, "example@example.com")
        self.assertEqual(event.reject_reason, "bounced")

    def test_dropped_unsubscribed_event(self):
        raw_events = [
            {
                "email": "unsubscribe@example.com",
                "anymail_id": "a36ec0f9-aabe-45c7-9a84-3e17afb5cb65",
                "timestamp": 1461095250,
                "smtp-id": "<wrfRRvF7Q0GgwUo2CvDmEA@ismtpd0006p1sjc2.sendgrid.net>",
                "sg_event_id": "oxy9OLwMTAy5EsuZn1qhIg",
                "sg_message_id": "filter0199p1las1.4745.5717FB6F5.0",
                "reason": "Unsubscribed Address",
                "event": "dropped",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "a36ec0f9-aabe-45c7-9a84-3e17afb5cb65")
        self.assertEqual(event.event_id, "oxy9OLwMTAy5EsuZn1qhIg")
        self.assertEqual(event.recipient, "unsubscribe@example.com")
        self.assertEqual(event.reject_reason, "unsubscribed")
        self.assertEqual(event.mta_response, None)

    def test_bounce_event(self):
        raw_events = [
            {
                "ip": "167.89.17.173",
                "status": "5.1.1",
                "smtp-id": "<wrfRRvF7Q0GgwUo2CvDmEA@ismtpd0006p1sjc2.sendgrid.net>",
                "sg_event_id": "lC0Rc-FuQmKbnxCWxX1jRQ",
                "reason": "550 5.1.1"
                " The email account that you tried to reach does not exist.",
                "sg_message_id": "Lli-03HcQ5-JLybO9fXsJg"
                ".filter0077p1las1.21536.5717FC482.0",
                "tls": 1,
                "event": "bounce",
                "email": "noreply@example.com",
                "timestamp": 1461095250,
                "anymail_id": "de212213-bb66-4302-8f3f-20acdb7a104e",
                "type": "bounce",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "de212213-bb66-4302-8f3f-20acdb7a104e")
        self.assertEqual(event.event_id, "lC0Rc-FuQmKbnxCWxX1jRQ")
        self.assertEqual(event.recipient, "noreply@example.com")
        self.assertEqual(
            event.mta_response,
            "550 5.1.1 The email account that you tried to reach does not exist.",
        )

    def test_deferred_event(self):
        raw_events = [
            {
                "response": "Email was deferred due to the following reason(s):"
                " [IPs were throttled by recipient server]",
                "smtp-id": "<wrfRRvF7Q0GgwUo2CvDmEA@ismtpd0006p1sjc2.sendgrid.net>",
                "sg_event_id": "b_syL5UiTvWC_Ky5L6Bs5Q",
                "sg_message_id": "u9Gvi3mzT6iC2poAb58_qQ.filter0465p1mdw1"
                ".8054.5718271B40.0",
                "event": "deferred",
                "email": "recipient@example.com",
                "attempt": "1",
                "timestamp": 1461200990,
                "anymail_id": "ccf83222-0d7e-4542-8beb-893122afa757",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "deferred")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "ccf83222-0d7e-4542-8beb-893122afa757")
        self.assertEqual(event.event_id, "b_syL5UiTvWC_Ky5L6Bs5Q")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.mta_response,
            "Email was deferred due to the following reason(s):"
            " [IPs were throttled by recipient server]",
        )

    def test_open_event(self):
        raw_events = [
            {
                "email": "recipient@example.com",
                "timestamp": 1461095250,
                "ip": "66.102.6.229",
                "sg_event_id": "MjIwNDg5NTgtZGE3OC00NDI1LWFiMmMtMDUyZTU2ZmFkOTFm",
                "sg_message_id": "wrfRRvF7Q0GgwUo2CvDmEA"
                ".filter0425p1mdw1.13037.57168B4A1D.0",
                "anymail_id": "44920b35-3e31-478b-bb67-b4f5e0c85ebc",
                "useragent": "Mozilla/5.0 (Windows NT 5.1; rv:11.0) Gecko Firefox/11.0",
                "event": "open",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "44920b35-3e31-478b-bb67-b4f5e0c85ebc")
        self.assertEqual(
            event.event_id, "MjIwNDg5NTgtZGE3OC00NDI1LWFiMmMtMDUyZTU2ZmFkOTFm"
        )
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent, "Mozilla/5.0 (Windows NT 5.1; rv:11.0) Gecko Firefox/11.0"
        )

    def test_click_event(self):
        raw_events = [
            {
                "ip": "24.130.34.103",
                "sg_event_id": "OTdlOGUzYjctYjc5Zi00OWE4LWE4YWUtNjIxNjk2ZTJlNGVi",
                "sg_message_id": "_fjPjuJfRW-IPs5SuvYotg"
                ".filter0590p1mdw1.2098.57168CFC4B.0",
                "useragent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_4)"
                " AppleWebKit/537.36",
                "anymail_id": "75de5af9-a090-4325-87f9-8c599ad66f60",
                "event": "click",
                "url_offset": {"index": 0, "type": "html"},
                "email": "recipient@example.com",
                "timestamp": 1461095250,
                "url": "http://www.example.com",
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "75de5af9-a090-4325-87f9-8c599ad66f60")
        self.assertEqual(
            event.event_id, "OTdlOGUzYjctYjc5Zi00OWE4LWE4YWUtNjIxNjk2ZTJlNGVi"
        )
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent,
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_4) AppleWebKit/537.36",
        )
        self.assertEqual(event.click_url, "http://www.example.com")

    def test_compatibility_message_id_from_smtp_id(self):
        # Prior to v3.0, Anymail tried to use a custom Message-ID header as
        # the `message_id`, and relied on SendGrid passing that to webhooks as
        # 'smtp-id'. Make sure webhooks extract message_id for messages sent
        # with earlier Anymail versions. (See issue #108.)
        raw_events = [
            {
                "ip": "167.89.17.173",
                "response": "250 2.0.0 OK 1461095248 m143si2210036ioe.159 - gsmtp ",
                "smtp-id": "<152712433591.85282.8340115595767222398@example.com>",
                "sg_event_id": "nOSv8m0eTQ-vxvwNwt3fZQ",
                "sg_message_id": "wrfRRvF7Q0GgwUo2CvDmEA.filter0425p1mdw1"
                ".13037.57168B4A1D.0",
                "tls": 1,
                "event": "delivered",
                "email": "recipient@example.com",
                "timestamp": 1461095250,
            }
        ]
        response = self.client.post(
            "/anymail/sendgrid/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendGridTrackingWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(
            event.message_id, "<152712433591.85282.8340115595767222398@example.com>"
        )
        self.assertEqual(event.metadata, {})  # smtp-id not left in metadata



import hashlib
import hmac
import json
from base64 import b64encode
from datetime import datetime, timezone
from unittest.mock import ANY
from urllib.parse import urljoin

from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings, tag

from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.mandrill import (
    MandrillCombinedWebhookView,
    MandrillTrackingWebhookView,
)

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase

TEST_WEBHOOK_KEY = "TEST_WEBHOOK_KEY"


def mandrill_args(
    events=None,
    host="http://testserver/",  # Django test-client default
    path="/anymail/mandrill/",  # Anymail urlconf default
    auth="username:password",  # WebhookTestCase default
    key=TEST_WEBHOOK_KEY,
):
    """Returns TestClient.post kwargs for Mandrill webhook call with events

    Computes correct signature.
    """
    if events is None:
        events = []
    test_client_path = urljoin(host, path)  # https://testserver/anymail/mandrill/
    if auth:
        # we can get away with this simplification in these controlled tests,
        # but don't ever construct urls like this in production code -- it's not safe!
        full_url = test_client_path.replace("://", "://" + auth + "@")
    else:
        full_url = test_client_path
    mandrill_events = json.dumps(events)
    signed_data = full_url + "mandrill_events" + mandrill_events
    signature = b64encode(
        hmac.new(
            key=key.encode("ascii"),
            msg=signed_data.encode("utf-8"),
            digestmod=hashlib.sha1,
        ).digest()
    )
    return {
        "path": test_client_path,
        "data": {"mandrill_events": mandrill_events},
        "HTTP_X_MANDRILL_SIGNATURE": signature,
    }


@tag("mandrill")
class MandrillWebhookSettingsTestCase(WebhookTestCase):
    def test_requires_webhook_key(self):
        with self.assertRaisesRegex(ImproperlyConfigured, r"MANDRILL_WEBHOOK_KEY"):
            self.client.post("/anymail/mandrill/", data={"mandrill_events": "[]"})

    def test_head_does_not_require_webhook_key(self):
        # Mandrill issues an unsigned HEAD request to verify the wehbook url.
        # Only *after* that succeeds will Mandrill will tell you the webhook key.
        # So make sure that HEAD request will go through without any key set:
        response = self.client.head("/anymail/mandrill/")
        self.assertEqual(response.status_code, 200)


@tag("mandrill")
@override_settings(ANYMAIL_MANDRILL_WEBHOOK_KEY=TEST_WEBHOOK_KEY)
class MandrillWebhookSecurityTestCase(WebhookBasicAuthTestCase):
    should_warn_if_no_auth = False  # because we check webhook signature

    def call_webhook(self):
        kwargs = mandrill_args([{"event": "send"}])
        return self.client.post(**kwargs)

    # Additional tests are in WebhookBasicAuthTestCase

    def test_verifies_correct_signature(self):
        kwargs = mandrill_args([{"event": "send"}])
        response = self.client.post(**kwargs)
        self.assertEqual(response.status_code, 200)

    def test_verifies_missing_signature(self):
        response = self.client.post(
            "/anymail/mandrill/", data={"mandrill_events": '[{"event":"send"}]'}
        )
        self.assertEqual(response.status_code, 400)

    def test_verifies_bad_signature(self):
        kwargs = mandrill_args([{"event": "send"}], key="wrong API key")
        response = self.client.post(**kwargs)
        self.assertEqual(response.status_code, 400)

    @override_settings(ANYMAIL={})  # clear WEBHOOK_SECRET from WebhookTestCase
    def test_no_basic_auth(self):
        # Signature validation should work properly if you're not using basic auth
        self.clear_basic_auth()
        kwargs = mandrill_args([{"event": "send"}], auth="")
        response = self.client.post(**kwargs)
        self.assertEqual(response.status_code, 200)

    @override_settings(
        ALLOWED_HOSTS=["127.0.0.1", ".example.com"],
        ANYMAIL={
            "MANDRILL_WEBHOOK_URL": "https://abcde:12345@example.com/anymail/mandrill/",
            "WEBHOOK_SECRET": "abcde:12345",
        },
    )
    def test_webhook_url_setting(self):
        # If Django can't build_absolute_uri correctly (e.g., because your proxy
        # frontend isn't setting the proxy headers correctly), you must set
        # MANDRILL_WEBHOOK_URL to the actual public url where Mandrill calls
        # the webhook.
        self.set_basic_auth("abcde", "12345")
        kwargs = mandrill_args(
            [{"event": "send"}], host="https://example.com/", auth="abcde:12345"
        )
        response = self.client.post(SERVER_NAME="127.0.0.1", **kwargs)
        self.assertEqual(response.status_code, 200)

    # override WebhookBasicAuthTestCase version of this test
    @override_settings(ANYMAIL={"WEBHOOK_SECRET": ["cred1:pass1", "cred2:pass2"]})
    def test_supports_credential_rotation(self):
        """You can supply a list of basic auth credentials, and any is allowed"""
        self.set_basic_auth("cred1", "pass1")
        response = self.client.post(**mandrill_args(auth="cred1:pass1"))
        self.assertEqual(response.status_code, 200)

        self.set_basic_auth("cred2", "pass2")
        response = self.client.post(**mandrill_args(auth="cred2:pass2"))
        self.assertEqual(response.status_code, 200)

        self.set_basic_auth("baduser", "wrongpassword")
        response = self.client.post(**mandrill_args(auth="baduser:wrongpassword"))
        self.assertEqual(response.status_code, 400)


@tag("mandrill")
@override_settings(ANYMAIL_MANDRILL_WEBHOOK_KEY=TEST_WEBHOOK_KEY)
class MandrillTrackingTestCase(WebhookTestCase):
    def test_head_request(self):
        # Mandrill verifies webhooks at config time with a HEAD request
        # (See MandrillWebhookSettingsTestCase above for equivalent
        # without the key yet set)
        response = self.client.head("/anymail/mandrill/tracking/")
        self.assertEqual(response.status_code, 200)

    def test_post_request_invalid_json(self):
        kwargs = mandrill_args()
        kwargs["data"] = {"mandrill_events": "GARBAGE DATA"}
        response = self.client.post(**kwargs)
        self.assertEqual(response.status_code, 400)

    def test_send_event(self):
        raw_events = [
            {
                "event": "send",
                "msg": {
                    "ts": 1461095211,  # time send called
                    "subject": "Webhook Test",
                    "email": "recipient@example.com",
                    "sender": "sender@example.com",
                    "tags": ["tag1", "tag2"],
                    "metadata": {"custom1": "value1", "custom2": "value2"},
                    "_id": "abcdef012345789abcdef012345789",
                },
                "_id": "abcdef012345789abcdef012345789",
                "ts": 1461095246,  # time of event
            }
        ]
        response = self.client.post(**mandrill_args(events=raw_events))
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MandrillCombinedWebhookView,
            event=ANY,
            esp_name="Mandrill",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "sent")
        self.assertEqual(
            event.timestamp, datetime(2016, 4, 19, 19, 47, 26, tzinfo=timezone.utc)
        )
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "abcdef012345789abcdef012345789")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.tags, ["tag1", "tag2"])
        self.assertEqual(event.metadata, {"custom1": "value1", "custom2": "value2"})

    def test_hard_bounce_event(self):
        raw_events = [
            {
                "event": "hard_bounce",
                "msg": {
                    "ts": 1461095211,  # time send called
                    "subject": "Webhook Test",
                    "email": "bounce@example.com",
                    "sender": "sender@example.com",
                    "bounce_description": "bad_mailbox",
                    "bgtools_code": 10,
                    "diag": "smtp;550 5.1.1 The email account that you tried"
                    " to reach does not exist.",
                    "_id": "abcdef012345789abcdef012345789",
                },
                "_id": "abcdef012345789abcdef012345789",
                "ts": 1461095246,  # time of event
            }
        ]
        response = self.client.post(**mandrill_args(events=raw_events))
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MandrillCombinedWebhookView,
            event=ANY,
            esp_name="Mandrill",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "abcdef012345789abcdef012345789")
        self.assertEqual(event.recipient, "bounce@example.com")
        self.assertEqual(
            event.mta_response,
            "smtp;550 5.1.1 The email account that you tried to reach does not exist.",
        )

    def test_click_event(self):
        raw_events = [
            {
                "event": "click",
                "msg": {
                    "ts": 1461095211,  # time send called
                    "subject": "Webhook Test",
                    "email": "recipient@example.com",
                    "sender": "sender@example.com",
                    "opens": [{"ts": 1461095242}],
                    "clicks": [{"ts": 1461095246, "url": "http://example.com"}],
                    "_id": "abcdef012345789abcdef012345789",
                },
                "user_agent": "Mozilla/5.0 (Windows NT 5.1; rv:11.0)"
                " Gecko Firefox/11.0",
                "url": "http://example.com",
                "_id": "abcdef012345789abcdef012345789",
                "ts": 1461095246,  # time of event
            }
        ]
        response = self.client.post(**mandrill_args(events=raw_events))
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MandrillCombinedWebhookView,
            event=ANY,
            esp_name="Mandrill",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.click_url, "http://example.com")
        self.assertEqual(
            event.user_agent, "Mozilla/5.0 (Windows NT 5.1; rv:11.0) Gecko Firefox/11.0"
        )

    def test_sync_event(self):
        # Mandrill sync events use a different format from other events
        # https://mandrill.zendesk.com/hc/en-us/articles/205583297-Sync-Event-Webhook-format
        raw_events = [
            {
                "type": "blacklist",
                "action": "add",
                "reject": {"email": "recipient@example.com", "reason": "manual edit"},
            }
        ]
        response = self.client.post(**mandrill_args(events=raw_events))
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MandrillCombinedWebhookView,
            event=ANY,
            esp_name="Mandrill",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "unknown")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.description, "manual edit")

    def test_old_tracking_url(self):
        # Earlier versions of Anymail used /mandrill/tracking/ (and didn't support
        # inbound); make sure that URL continues to work.
        raw_events = [
            {
                "event": "send",
                "msg": {
                    "ts": 1461095211,  # time send called
                    "subject": "Webhook Test",
                    "email": "recipient@example.com",
                    "sender": "sender@example.com",
                    "tags": ["tag1", "tag2"],
                    "metadata": {"custom1": "value1", "custom2": "value2"},
                    "_id": "abcdef012345789abcdef012345789",
                },
                "_id": "abcdef012345789abcdef012345789",
                "ts": 1461095246,  # time of event
            }
        ]
        response = self.client.post(
            **mandrill_args(events=raw_events, path="/anymail/mandrill/tracking/")
        )
        self.assertEqual(response.status_code, 200)
        self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MandrillTrackingWebhookView,
            event=ANY,
            esp_name="Mandrill",
        )



import json
from io import BytesIO
from unittest.mock import patch

import requests
from django.core import mail
from django.test import SimpleTestCase

from anymail.exceptions import AnymailAPIError

from .utils import AnymailTestMixin

UNSET = object()


class RequestsBackendMockAPITestCase(AnymailTestMixin, SimpleTestCase):
    """TestCase that mocks API calls through requests"""

    DEFAULT_RAW_RESPONSE = b"""{"subclass": "should override"}"""
    DEFAULT_CONTENT_TYPE = None  # e.g., "application/json"
    DEFAULT_STATUS_CODE = 200  # most APIs use '200 OK' for success

    class MockResponse(requests.Response):
        """requests.request return value mock sufficient for testing"""

        def __init__(
            self,
            status_code=200,
            raw=b"RESPONSE",
            content_type=None,
            encoding="utf-8",
            reason=None,
            test_case=None,
        ):
            super().__init__()
            self.status_code = status_code
            self.encoding = encoding
            self.reason = reason or ("OK" if 200 <= status_code < 300 else "ERROR")
            self.raw = BytesIO(raw)
            if content_type is not None:
                self.headers["Content-Type"] = content_type
            self.test_case = test_case

        @property
        def url(self):
            return self.test_case.get_api_call_arg("url", required=False)

        @url.setter
        def url(self, url):
            if url is not None:
                raise ValueError("MockResponse can't handle url assignment")

    def setUp(self):
        super().setUp()
        self.patch_request = patch("requests.Session.request", autospec=True)
        self.mock_request = self.patch_request.start()
        self.addCleanup(self.patch_request.stop)
        self.set_mock_response()

    def set_mock_response(
        self,
        status_code=UNSET,
        raw=UNSET,
        json_data=UNSET,
        encoding="utf-8",
        content_type=UNSET,
        reason=None,
    ):
        if status_code is UNSET:
            status_code = self.DEFAULT_STATUS_CODE
        if json_data is not UNSET:
            assert raw is UNSET, "provide json_data or raw, not both"
            raw = json.dumps(json_data).encode(encoding)
            if content_type is UNSET:
                content_type = "application/json"
        if raw is UNSET:
            raw = self.DEFAULT_RAW_RESPONSE
        if content_type is UNSET:
            content_type = self.DEFAULT_CONTENT_TYPE
        mock_response = self.MockResponse(
            status_code,
            raw=raw,
            content_type=content_type,
            encoding=encoding,
            reason=reason,
            test_case=self,
        )
        self.mock_request.return_value = mock_response
        return mock_response

    def assert_esp_called(self, url, method="POST"):
        """Verifies the (mock) ESP API was called on endpoint.

        url can be partial, and is just checked against the end of the url requested
        """
        # This assumes the last (or only) call to requests.Session.request
        # is the API call of interest.
        if self.mock_request.call_args is None:
            raise AssertionError("No ESP API was called")
        if method is not None:
            actual_method = self.get_api_call_arg("method")
            if actual_method != method:
                self.fail(
                    "API was not called using %s. (%s was used instead.)"
                    % (method, actual_method)
                )
        if url is not None:
            actual_url = self.get_api_call_arg("url")
            if not actual_url.endswith(url):
                self.fail(
                    "API was not called at %s\n(It was called at %s)"
                    % (url, actual_url)
                )

    def get_api_call_arg(self, kwarg, required=True):
        """Returns an argument passed to the mock ESP API.

        Fails test if API wasn't called.
        """
        if self.mock_request.call_args is None:
            raise AssertionError("API was not called")
        (args, kwargs) = self.mock_request.call_args
        try:
            return kwargs[kwarg]
        except KeyError:
            pass

        try:
            # positional arg? This is the order of requests.Session.request params:
            pos = (
                "method",
                "url",
                "params",
                "data",
                "headers",
                "cookies",
                "files",
                "auth",
                "timeout",
                "allow_redirects",
                "proxies",
                "hooks",
                "stream",
                "verify",
                "cert",
                "json",
            ).index(kwarg)
            return args[pos]
        except (ValueError, IndexError):
            pass

        if required:
            self.fail("API was called without required arg '%s'" % kwarg)
        return None

    def get_api_call_params(self, required=True):
        """Returns the query params sent to the mock ESP API."""
        return self.get_api_call_arg("params", required)

    def get_api_call_data(self, required=True):
        """Returns the raw data sent to the mock ESP API."""
        return self.get_api_call_arg("data", required)

    def get_api_call_json(self, required=True):
        """Returns the data sent to the mock ESP API, json-parsed"""
        # could be either the data param (as json str)
        # or the json param (needing formatting)
        value = self.get_api_call_arg("data", required=False)
        if value is not None:
            return json.loads(value)
        else:
            return self.get_api_call_arg("json", required)

    def get_api_call_headers(self, required=True):
        """Returns the headers sent to the mock ESP API"""
        return self.get_api_call_arg("headers", required)

    def get_api_call_files(self, required=True):
        """Returns the files sent to the mock ESP API"""
        return self.get_api_call_arg("files", required)

    def get_api_call_auth(self, required=True):
        """Returns the auth sent to the mock ESP API"""
        return self.get_api_call_arg("auth", required)

    def get_api_prepared_request(self):
        """Returns the PreparedRequest that would have been sent"""
        (args, kwargs) = self.mock_request.call_args
        kwargs.pop("timeout", None)  # Session-only param
        request = requests.Request(**kwargs)
        return request.prepare()

    def assert_esp_not_called(self, msg=None):
        if self.mock_request.called:
            raise AssertionError(msg or "ESP API was called and shouldn't have been")


class SessionSharingTestCases(RequestsBackendMockAPITestCase):
    """Common test cases for requests backend connection sharing.

    Instantiate for each ESP by:
    - subclassing
    - adding or overriding any tests as appropriate
    """

    def __init__(self, methodName="runTest"):
        if self.__class__ is SessionSharingTestCases:
            # don't run these tests on the abstract base implementation
            methodName = "runNoTestsInBaseClass"
        super().__init__(methodName)

    def runNoTestsInBaseClass(self):
        pass

    def setUp(self):
        super().setUp()
        self.patch_close = patch("requests.Session.close", autospec=True)
        self.mock_close = self.patch_close.start()
        self.addCleanup(self.patch_close.stop)

    def test_connection_sharing(self):
        """RequestsBackend reuses one requests session when sending multiple messages"""
        datatuple = (
            ("Subject 1", "Body 1", "from@example.com", ["to@example.com"]),
            ("Subject 2", "Body 2", "from@example.com", ["to@example.com"]),
        )
        mail.send_mass_mail(datatuple)
        self.assertEqual(self.mock_request.call_count, 2)
        session1 = self.mock_request.call_args_list[0][0]  # arg[0] (self) is session
        session2 = self.mock_request.call_args_list[1][0]
        self.assertEqual(session1, session2)
        self.assertEqual(self.mock_close.call_count, 1)

    def test_caller_managed_connections(self):
        """Calling code can created long-lived connection that it opens and closes"""
        connection = mail.get_connection()
        connection.open()
        mail.send_mail(
            "Subject 1",
            "body",
            "from@example.com",
            ["to@example.com"],
            connection=connection,
        )
        session1 = self.mock_request.call_args[0]
        self.assertEqual(self.mock_close.call_count, 0)  # shouldn't be closed yet

        mail.send_mail(
            "Subject 2",
            "body",
            "from@example.com",
            ["to@example.com"],
            connection=connection,
        )
        self.assertEqual(self.mock_close.call_count, 0)  # still shouldn't be closed
        session2 = self.mock_request.call_args[0]
        self.assertEqual(session1, session2)  # should have reused same session

        connection.close()
        self.assertEqual(self.mock_close.call_count, 1)

    def test_session_closed_after_exception(self):
        self.set_mock_response(status_code=500)
        with self.assertRaises(AnymailAPIError):
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        self.assertEqual(self.mock_close.call_count, 1)

    def test_session_closed_after_fail_silently_exception(self):
        self.set_mock_response(status_code=500)
        sent = mail.send_mail(
            "Subject",
            "Message",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)
        self.assertEqual(self.mock_close.call_count, 1)

    def test_caller_managed_session_closed_after_exception(self):
        connection = mail.get_connection()
        connection.open()
        self.set_mock_response(status_code=500)
        with self.assertRaises(AnymailAPIError):
            mail.send_mail(
                "Subject",
                "Message",
                "from@example.com",
                ["to@example.com"],
                connection=connection,
            )
        self.assertEqual(self.mock_close.call_count, 0)  # wait for us to close it

        connection.close()
        self.assertEqual(self.mock_close.call_count, 1)



import base64
import json
from datetime import datetime, timezone
from unittest import skipIf, skipUnless
from unittest.mock import ANY

from django.test import override_settings, tag

from anymail.exceptions import AnymailImproperlyInstalled, AnymailInsecureWebhookWarning
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.resend import ResendTrackingWebhookView

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase

# These tests are run both with and without 'svix' installed.
try:
    from svix import Webhook
except ImportError:
    SVIX_INSTALLED = False
    Webhook = None
else:
    SVIX_INSTALLED = True


def svix_secret(secret):
    return f"whsec_{base64.b64encode(secret.encode('ascii')).decode('ascii')}"


TEST_SIGNING_SECRET = svix_secret("TEST_SIGNING_SECRET") if SVIX_INSTALLED else None
TEST_WEBHOOK_MESSAGE_ID = "msg_abcdefghijklmnopqrst12345"


class ResendWebhookTestCase(WebhookTestCase):
    def client_post_signed(self, url, json_data, svix_id=None, secret=None):
        """Return self.client.post(url, serialized json_data) signed with secret"""
        svix_id = svix_id or TEST_WEBHOOK_MESSAGE_ID
        secret = secret or TEST_SIGNING_SECRET
        data = json.dumps(json_data)
        headers = {
            "svix-id": svix_id,
        }

        if SVIX_INSTALLED:
            timestamp = datetime.now(tz=timezone.utc)
            signature = Webhook(secret).sign(
                msg_id=svix_id, timestamp=timestamp, data=data
            )
            headers.update(
                {
                    "svix-timestamp": timestamp.timestamp(),
                    "svix-signature": signature,
                }
            )

        return self.client.post(
            url,
            content_type="application/json",
            data=data.encode("utf-8"),
            # Django 4.2+ test Client allows headers=headers;
            # before that, must convert to HTTP_ args:
            **{
                f"HTTP_{header.upper().replace('-', '_')}": value
                for header, value in headers.items()
            },
        )


@tag("resend")
@override_settings(ANYMAIL={})  # clear WEBHOOK_SECRET from base class
class ResendWebhookSettingsTestCase(ResendWebhookTestCase):
    @skipIf(SVIX_INSTALLED, "test covers behavior when 'svix' package missing")
    @override_settings(ANYMAIL_RESEND_SIGNING_SECRET=svix_secret("settings secret"))
    def test_secret_requires_svix_installed(self):
        """If webhook secret is specified, error if svix not available to verify"""
        with self.assertRaisesMessage(AnymailImproperlyInstalled, "svix"):
            self.client_post_signed("/anymail/resend/tracking/", {"type": "email.sent"})

    # Test with and without SVIX_INSTALLED
    def test_basic_auth_required_without_secret(self):
        with self.assertWarns(AnymailInsecureWebhookWarning):
            self.client_post_signed("/anymail/resend/tracking/", {"type": "email.sent"})

    # Test with and without SVIX_INSTALLED
    @override_settings(ANYMAIL={"WEBHOOK_SECRET": "username:password"})
    def test_signing_secret_optional_with_basic_auth(self):
        """Secret verification is optional if using basic auth"""
        response = self.client_post_signed(
            "/anymail/resend/tracking/", {"type": "email.sent"}
        )
        self.assertEqual(response.status_code, 200)

    @skipUnless(SVIX_INSTALLED, "secret verification requires 'svix' package")
    @override_settings(ANYMAIL_RESEND_SIGNING_SECRET=svix_secret("settings secret"))
    def test_signing_secret_view_params(self):
        """Webhook signing secret can be provided as a view param"""
        view_secret = svix_secret("view-level secret")
        view = ResendTrackingWebhookView.as_view(signing_secret=view_secret)
        view_instance = view.view_class(**view.view_initkwargs)
        self.assertEqual(view_instance.signing_secret, view_secret)


@tag("resend")
@override_settings(ANYMAIL_RESEND_SIGNING_SECRET=TEST_SIGNING_SECRET)
class ResendWebhookSecurityTestCase(ResendWebhookTestCase, WebhookBasicAuthTestCase):
    should_warn_if_no_auth = TEST_SIGNING_SECRET is None

    def call_webhook(self):
        return self.client_post_signed(
            "/anymail/resend/tracking/",
            {"type": "email.sent"},
            secret=TEST_SIGNING_SECRET,
        )

    # Additional tests are in WebhookBasicAuthTestCase

    @skipUnless(SVIX_INSTALLED, "signature verification requires 'svix' package")
    def test_verifies_correct_signature(self):
        response = self.client_post_signed(
            "/anymail/resend/tracking/",
            {"type": "email.sent"},
            secret=TEST_SIGNING_SECRET,
        )
        self.assertEqual(response.status_code, 200)

    @skipUnless(SVIX_INSTALLED, "signature verification requires 'svix' package")
    def test_verifies_missing_signature(self):
        response = self.client.post(
            "/anymail/resend/tracking/",
            content_type="application/json",
            data={"type": "email.sent"},
        )
        self.assertEqual(response.status_code, 400)

    @skipUnless(SVIX_INSTALLED, "signature verification requires 'svix' package")
    def test_verifies_bad_signature(self):
        # This also verifies that the error log references the correct setting to check.
        with self.assertLogs() as logs:
            response = self.client_post_signed(
                "/anymail/resend/tracking/",
                {"type": "email.sent"},
                secret=svix_secret("wrong signing key"),
            )
        # SuspiciousOperation causes 400 response (even in test client):
        self.assertEqual(response.status_code, 400)
        self.assertIn("check Anymail RESEND_SIGNING_SECRET", logs.output[0])


@tag("resend")
@override_settings(ANYMAIL_RESEND_SIGNING_SECRET=TEST_SIGNING_SECRET)
class ResendTestCase(ResendWebhookTestCase):
    def test_sent_event(self):
        raw_event = {
            "created_at": "2023-09-28T17:19:43.736Z",
            "data": {
                "created_at": "2023-09-28T17:19:43.982Z",
                "email_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "from": "Sender <from@example.com>",
                "headers": [
                    {"name": "Reply-To", "value": "reply@example.com"},
                    {"name": "X-Tags", "value": '["tag1", "Tag 2"]'},
                    {
                        "name": "X-Metadata",
                        "value": '{"cohort": "2018-08-B", "user_id": 123456}',
                    },
                    {"name": "Cc", "value": "cc1@example.org, Cc 2 <cc2@example.org>"},
                ],
                "subject": "Sending test",
                "tags": {"tag1": "Tag_1_value", "tag2": "Tag_2_value"},
                "to": ["Recipient <to@example.org>", "to2@example.org"],
            },
            "type": "email.sent",
        }
        response = self.client_post_signed(
            "/anymail/resend/tracking/",
            raw_event,
            svix_id="msg_2W2D3qXLS5fOaPja1GDg7rF2CwB",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=ResendTrackingWebhookView,
            event=ANY,
            esp_name="Resend",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "sent")
        # event.timestamp comes from root-level created_at:
        self.assertEqual(
            event.timestamp,
            # "2023-09-28T17:19:43.736Z"
            datetime(2023, 9, 28, 17, 19, 43, microsecond=736000, tzinfo=timezone.utc),
        )
        # event.message_id matches the message.anymail_status.message_id when the
        # message was sent. It comes from data.email_id:
        self.assertEqual(event.message_id, "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")
        # event.event_id is unique for each event, and comes from svix-id header:
        self.assertEqual(event.event_id, "msg_2W2D3qXLS5fOaPja1GDg7rF2CwB")
        # event.recipient is always the first "to" addr:
        self.assertEqual(event.recipient, "to@example.org")
        self.assertEqual(event.tags, ["tag1", "Tag 2"])
        self.assertEqual(event.metadata, {"cohort": "2018-08-B", "user_id": 123456})
        self.assertEqual(event.esp_event, raw_event)

        # You can retrieve Resend native tags (which are different from Anymail tags)
        # from esp_event:
        resend_tags = event.esp_event["data"].get("tags", {})
        self.assertEqual(resend_tags, {"tag1": "Tag_1_value", "tag2": "Tag_2_value"})

    def test_delivered_event(self):
        raw_event = {
            "created_at": "2023-09-28T17:19:44.823Z",
            "data": {
                "created_at": "2023-09-28T17:19:43.982Z",
                "email_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "from": "Sender <from@example.com>",
                "subject": "Sending test",
                "to": ["to@example.org"],
            },
            "type": "email.delivered",
        }
        response = self.client_post_signed("/anymail/resend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=ResendTrackingWebhookView,
            event=ANY,
            esp_name="Resend",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(event.recipient, "to@example.org")
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_hard_bounced_event(self):
        raw_event = {
            "created_at": "2023-10-02T18:11:26.101Z",
            "data": {
                "bounce": {
                    "message": (
                        "The recipient's email provider sent a hard bounce message, but"
                        " didn't specify the reason for the hard bounce. We recommend"
                        " removing the recipient's email address from your mailing list."
                        " Sending messages to addresses that produce hard bounces can"
                        " have a negative impact on your reputation as a sender."
                    )
                },
                "created_at": "2023-10-02T18:11:25.729Z",
                "email_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "from": "Sender <from@example.com>",
                "subject": "Sending test",
                "to": ["bounced@resend.dev"],
            },
            "type": "email.bounced",
        }
        response = self.client_post_signed("/anymail/resend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=ResendTrackingWebhookView,
            event=ANY,
            esp_name="Resend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertRegex(
            event.description,
            r"^The recipient's email provider sent a hard bounce message.*",
        )
        self.assertIsNone(event.mta_response)  # raw MTA info not provided

    def test_suppressed_event(self):
        raw_event = {
            "created_at": "2023-10-01T20:01:01.598Z",
            "data": {
                "bounce": {
                    "message": (
                        "Resend has suppressed sending to this address because it is"
                        " on the account-level suppression list. This does not count"
                        " toward your bounce rate metric"
                    )
                },
                "created_at": "2023-10-01T20:01:01.339Z",
                "email_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "from": "Sender <from@example.com>",
                "subject": "Sending test",
                "to": ["blocked@example.org"],
            },
            "type": "email.bounced",
        }
        response = self.client_post_signed("/anymail/resend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=ResendTrackingWebhookView,
            event=ANY,
            esp_name="Resend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.reject_reason, "blocked")
        self.assertRegex(
            event.description, r"^Resend has suppressed sending to this address.*"
        )
        self.assertIsNone(event.mta_response)  # raw MTA info not provided

    def test_delivery_delayed_event(self):
        # Haven't been able to trigger a real-world version of this event
        # (even with SMTP reply 450, status 4.0.0 "temporary failure").
        # This is the sample payload from Resend's docs, but correcting the type
        # from "email.delivered_delayed" to "email.delivery_delayed" to match
        # docs and configuration UI.
        raw_event = {
            "type": "email.delivery_delayed",  # "email.delivered_delayed",
            "created_at": "2023-02-22T23:41:12.126Z",
            "data": {
                "created_at": "2023-02-22T23:41:11.894719+00:00",
                "email_id": "56761188-7520-42d8-8898-ff6fc54ce618",
                "from": "Acme <onboarding@resend.dev>",
                "to": ["delivered@resend.dev"],
                "subject": "Sending this example",
            },
        }
        response = self.client_post_signed("/anymail/resend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=ResendTrackingWebhookView,
            event=ANY,
            esp_name="Resend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "deferred")
        self.assertIsNone(event.reject_reason)
        self.assertIsNone(event.description)
        self.assertIsNone(event.mta_response)  # raw MTA info not provided

    def test_complained_event(self):
        raw_event = {
            "created_at": "2023-10-02T18:10:03.690Z",
            "data": {
                "created_at": "2023-10-02T18:10:03.241Z",
                "email_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "from": "Sender <from@example.com>",
                "subject": "Sending test",
                "to": ["complained@resend.dev"],
            },
            "type": "email.complained",
        }
        response = self.client_post_signed("/anymail/resend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=ResendTrackingWebhookView,
            event=ANY,
            esp_name="Resend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "complained")

    def test_opened_event(self):
        raw_event = {
            "created_at": "2023-09-28T17:20:38.990Z",
            "data": {
                "created_at": "2023-09-28T17:19:43.982Z",
                "email_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "from": "Sender <from@example.com>",
                "subject": "Sending test",
                "to": ["to@example.org"],
            },
            "type": "email.opened",
        }
        response = self.client_post_signed("/anymail/resend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=ResendTrackingWebhookView,
            event=ANY,
            esp_name="Resend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")

    def test_clicked_event(self):
        raw_event = {
            "created_at": "2023-09-28T17:21:35.257Z",
            "data": {
                "click": {
                    "ipAddress": "192.168.1.101",
                    "link": "https://example.com/test",
                    "timestamp": "2023-09-28T17:21:35.257Z",
                    "userAgent": "Mozilla/5.0 ...",
                },
                "created_at": "2023-09-28T17:19:43.982Z",
                "email_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "from": "Sender <from@example.com>",
                "subject": "Sending test",
                "to": ["to@example.org"],
            },
            "type": "email.clicked",
        }
        response = self.client_post_signed("/anymail/resend/tracking/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=ResendTrackingWebhookView,
            event=ANY,
            esp_name="Resend",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.click_url, "https://example.com/test")
        self.assertEqual(event.user_agent, "Mozilla/5.0 ...")



from unittest.mock import patch

from django.core.mail import EmailMultiAlternatives
from django.test import SimpleTestCase

from anymail.message import AnymailRecipientStatus, AnymailStatus, attach_inline_image

from .utils import AnymailTestMixin, sample_image_content


class InlineImageTests(AnymailTestMixin, SimpleTestCase):
    def setUp(self):
        self.message = EmailMultiAlternatives()
        super().setUp()

    @patch("socket.getfqdn")
    def test_default_domain(self, mock_getfqdn):
        """The default Content-ID domain should *not* use local hostname"""
        # (This avoids problems with ESPs that re-use Content-ID as attachment
        # filename: if the local hostname ends in ".com", you can end up with
        # an inline attachment filename that causes Gmail to reject the message.)
        mock_getfqdn.return_value = "server.example.com"
        cid = attach_inline_image(self.message, sample_image_content())
        self.assertRegex(
            cid,
            r"[\w.]+@inline",
            "Content-ID should be a valid Message-ID, " "but _not_ @server.example.com",
        )

    def test_domain_override(self):
        cid = attach_inline_image(
            self.message, sample_image_content(), domain="example.org"
        )
        self.assertRegex(
            cid,
            r"[\w.]+@example\.org",
            "Content-ID should be a valid Message-ID @example.org",
        )


class AnymailStatusTests(AnymailTestMixin, SimpleTestCase):
    def test_single_recipient(self):
        recipients = {
            "one@example.com": AnymailRecipientStatus("12345", "sent"),
        }
        status = AnymailStatus()
        status.set_recipient_status(recipients)
        self.assertEqual(status.status, {"sent"})
        self.assertEqual(status.message_id, "12345")
        self.assertEqual(status.recipients, recipients)
        self.assertEqual(
            repr(status),
            "AnymailStatus<status={'sent'}, message_id='12345', 1 recipients>",
        )
        self.assertEqual(
            repr(status.recipients["one@example.com"]),
            "AnymailRecipientStatus('12345', 'sent')",
        )

    def test_multiple_recipients(self):
        recipients = {
            "one@example.com": AnymailRecipientStatus("12345", "sent"),
            "two@example.com": AnymailRecipientStatus("45678", "queued"),
        }
        status = AnymailStatus()
        status.set_recipient_status(recipients)
        self.assertEqual(status.status, {"queued", "sent"})
        self.assertEqual(status.message_id, {"12345", "45678"})
        self.assertEqual(status.recipients, recipients)
        self.assertEqual(
            repr(status),
            "AnymailStatus<status={'queued', 'sent'},"
            " message_id={'12345', '45678'}, 2 recipients>",
        )

    def test_multiple_recipients_same_message_id(self):
        # status.message_id collapses when it's the same for all recipients
        recipients = {
            "one@example.com": AnymailRecipientStatus("12345", "sent"),
            "two@example.com": AnymailRecipientStatus("12345", "queued"),
        }
        status = AnymailStatus()
        status.set_recipient_status(recipients)
        self.assertEqual(status.message_id, "12345")
        self.assertEqual(
            repr(status),
            "AnymailStatus<status={'queued', 'sent'},"
            " message_id='12345', 2 recipients>",
        )

    def test_none(self):
        status = AnymailStatus()
        self.assertIsNone(status.status)
        self.assertIsNone(status.message_id)
        self.assertEqual(repr(status), "AnymailStatus<status=None>")

    def test_invalid_message_id(self):
        with self.assertRaisesMessage(TypeError, "Invalid message_id"):
            AnymailRecipientStatus(["id-list", "is-not-valid"], "queued")

    def test_invalid_status(self):
        with self.assertRaisesMessage(ValueError, "Invalid status"):
            AnymailRecipientStatus("12345", "not-a-known-status")



import json
from base64 import b64decode, b64encode
from datetime import date, datetime, timezone
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from django.core import mail
from django.test import SimpleTestCase, override_settings, tag
from django.utils.timezone import (
    get_fixed_timezone,
    override as override_current_timezone,
)

from anymail.exceptions import (
    AnymailAPIError,
    AnymailConfigurationError,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import AnymailMessage, attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    sample_image_content,
    sample_image_path,
)


@tag("brevo")
@override_settings(
    EMAIL_BACKEND="anymail.backends.brevo.EmailBackend",
    ANYMAIL={"BREVO_API_KEY": "test_api_key"},
)
class BrevoBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = (
        b'{"messageId":"<201801020304.1234567890@smtp-relay.mailin.fr>"}'
    )
    DEFAULT_STATUS_CODE = 201  # Brevo v3 uses '201 Created' for success (in most cases)

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("brevo")
class BrevoBackendStandardEmailTests(BrevoBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@sender.example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called("https://api.brevo.com/v3/smtp/email")
        http_headers = self.get_api_call_headers()
        self.assertEqual(http_headers["api-key"], "test_api_key")
        self.assertEqual(http_headers["Content-Type"], "application/json")

        data = self.get_api_call_json()
        self.assertEqual(data["subject"], "Subject here")
        self.assertEqual(data["textContent"], "Here is the message.")
        self.assertEqual(data["sender"], {"email": "from@sender.example.com"})
        self.assertEqual(data["to"], [{"email": "to@example.com"}])

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["sender"], {"email": "from@example.com", "name": "From Name"}
        )
        self.assertEqual(
            data["to"],
            [
                {"email": "to1@example.com", "name": "Recipient #1"},
                {"email": "to2@example.com"},
            ],
        )
        self.assertEqual(
            data["cc"],
            [
                {"email": "cc1@example.com", "name": "Carbon Copy"},
                {"email": "cc2@example.com"},
            ],
        )
        self.assertEqual(
            data["bcc"],
            [
                {"email": "bcc1@example.com", "name": "Blind Copy"},
                {"email": "bcc2@example.com"},
            ],
        )

    def test_email_message(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com", "Also To <to2@example.com>"],
            bcc=["bcc1@example.com", "Also BCC <bcc2@example.com>"],
            cc=["cc1@example.com", "Also CC <cc2@example.com>"],
            headers={
                "Reply-To": "another@example.com",
                "X-MyHeader": "my value",
                # should override backend msgid:
                "Message-ID": "<mycustommsgid@sales.example.com>",
            },
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["sender"], {"email": "from@example.com"})
        self.assertEqual(data["subject"], "Subject")
        self.assertEqual(data["textContent"], "Body goes here")
        self.assertEqual(data["replyTo"], {"email": "another@example.com"})
        self.assertEqual(
            data["headers"],
            {
                "X-MyHeader": "my value",
                "Message-ID": "<mycustommsgid@sales.example.com>",
            },
        )

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["textContent"], text_content)
        self.assertEqual(data["htmlContent"], html_content)

        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("attachments", data)

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["htmlContent"], html_content)
        self.assertNotIn("textContent", data)

    def test_extra_headers(self):
        self.message.extra_headers = {
            "X-Custom": "string",
            "X-Num": 123,
            "Reply-To": '"Do Not Reply" <noreply@example.com>',
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["headers"]["X-Custom"], "string")
        # Header values must be strings (changed 11/2022)
        self.assertEqual(data["headers"]["X-Num"], "123")
        # Reply-To must be moved to separate param
        self.assertNotIn("Reply-To", data["headers"])
        self.assertEqual(
            data["replyTo"], {"name": "Do Not Reply", "email": "noreply@example.com"}
        )

    def test_extra_headers_serialization_error(self):
        self.message.extra_headers = {"X-Custom": Decimal(12.5)}
        with self.assertRaisesMessage(AnymailSerializationError, "Decimal"):
            self.message.send()

    def test_reply_to(self):
        self.message.reply_to = ['"Reply recipient" <reply@example.com']
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["replyTo"], {"name": "Reply recipient", "email": "reply@example.com"}
        )

    def test_multiple_reply_to(self):
        # Brevo v3 only allows a single reply address
        self.message.reply_to = [
            '"Reply recipient" <reply@example.com',
            "reply2@example.com",
        ]
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    @override_settings(ANYMAIL_IGNORE_UNSUPPORTED_FEATURES=True)
    def test_multiple_reply_to_ignore_unsupported(self):
        # Should use first Reply-To if ignoring unsupported features
        self.message.reply_to = [
            '"Reply recipient" <reply@example.com',
            "reply2@example.com",
        ]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["replyTo"], {"name": "Reply recipient", "email": "reply@example.com"}
        )

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["attachment"]), 3)

        attachments = data["attachment"]
        self.assertEqual(
            attachments[0],
            {
                "name": "test.txt",
                "content": b64encode(text_content.encode("utf-8")).decode("ascii"),
            },
        )
        self.assertEqual(
            attachments[1],
            {"name": "test.png", "content": b64encode(png_content).decode("ascii")},
        )
        self.assertEqual(
            attachments[2],
            {"name": "", "content": b64encode(pdf_content).decode("ascii")},
        )

    def test_unicode_attachment_correctly_decoded(self):
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        attachment = self.get_api_call_json()["attachment"][0]
        self.assertEqual(attachment["name"], "Une pièce jointe.html")
        self.assertEqual(
            b64decode(attachment["content"]).decode("utf-8"), "<p>\u2019</p>"
        )

    def test_embedded_images(self):
        # Brevo doesn't support inline image
        # inline image are just added as a content attachment

        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly
        image = MIMEImage(image_data)
        self.message.attach(image)

        self.message.send()

        image_data_b64 = b64encode(image_data).decode("ascii")
        data = self.get_api_call_json()
        self.assertEqual(
            data["attachment"][0],
            {
                "name": image_filename,  # the named one
                "content": image_data_b64,
            },
        )
        self.assertEqual(
            data["attachment"][1],
            {
                "name": "",  # the unnamed one
                "content": image_data_b64,
            },
        )

    def test_multiple_html_alternatives(self):
        self.message.body = "Text body"
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative(
            "<p>And maybe second html, too</p>", "text/html"
        )

        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_non_html_alternative(self):
        self.message.body = "Text body"
        self.message.attach_alternative("{'maybe': 'allowed'}", "application/json")

        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_api_failure(self):
        self.set_mock_response(status_code=400)
        with self.assertRaisesMessage(AnymailAPIError, "Brevo API response 400"):
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=400)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        # JSON error response:
        error_response = b"""{
            "code": "invalid_parameter",
            "message": "valid sender email required"
        }"""
        self.set_mock_response(status_code=400, raw=error_response)
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertIn("code", str(err))
        self.assertIn("message", str(err))

        # No content in the error response:
        self.set_mock_response(status_code=502, raw=None)
        with self.assertRaises(AnymailAPIError):
            self.message.send()


@tag("brevo")
class BrevoBackendAnymailFeatureTests(BrevoBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        # Brevo does not have a way to change envelope sender.
        self.message.envelope_sender = "anything@bounces.example.com"
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "envelope_sender"):
            self.message.send()

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": 6, "float": 98.6}
        self.message.send()

        data = self.get_api_call_json()

        metadata = json.loads(data["headers"]["X-Mailin-custom"])
        self.assertEqual(metadata["user_id"], "12345")
        self.assertEqual(metadata["items"], 6)
        self.assertEqual(metadata["float"], 98.6)

    def test_send_at(self):
        utc_plus_6 = get_fixed_timezone(6 * 60)
        utc_minus_8 = get_fixed_timezone(-8 * 60)

        with override_current_timezone(utc_plus_6):
            # Timezone-aware datetime converted to UTC:
            self.message.send_at = datetime(
                2016, 3, 4, 5, 6, 7, 8000, tzinfo=utc_minus_8
            )
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduledAt"], "2016-03-04T05:06:07.008-08:00")

            # Explicit UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=timezone.utc)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduledAt"], "2016-03-04T05:06:07.000+00:00")

            # Timezone-naive datetime assumed to be Django current_timezone
            # (also checks stripping microseconds)
            self.message.send_at = datetime(2022, 10, 11, 12, 13, 14, 567)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduledAt"], "2022-10-11T12:13:14.000+06:00")

            # Date-only treated as midnight in current timezone
            self.message.send_at = date(2022, 10, 22)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduledAt"], "2022-10-22T00:00:00.000+06:00")

            # POSIX timestamp
            self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduledAt"], "2022-05-06T07:08:09.000+00:00")

            # String passed unchanged (this is *not* portable between ESPs)
            self.message.send_at = "2022-10-13T18:02:00.123-11:30"
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduledAt"], "2022-10-13T18:02:00.123-11:30")

    def test_tag(self):
        self.message.tags = ["receipt", "multiple"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["tags"], ["receipt", "multiple"])

    def test_tracking(self):
        # Test one way...
        self.message.track_clicks = False
        self.message.track_opens = True

        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

        # ...and the opposite way
        self.message.track_clicks = True
        self.message.track_opens = False
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_template_id(self):
        # subject, body, and from_email must be None for Brevo template send:
        message = mail.EmailMessage(
            subject="My Subject",
            body=None,
            from_email="from@example.com",
            # single 'to' recommended (all 'to' get the same message)
            to=["Recipient <to@example.com>"],
            cc=["Recipient <cc1@example.com>", "Recipient <cc2@example.com>"],
            bcc=["Recipient <bcc@example.com>"],
            reply_to=["Recipient <reply@example.com>"],
        )
        # Brevo uses per-account numeric ID to identify templates:
        message.template_id = 12
        message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["templateId"], 12)
        self.assertEqual(data["subject"], "My Subject")
        self.assertEqual(data["to"], [{"email": "to@example.com", "name": "Recipient"}])

    _mock_batch_response = {
        "messageIds": [
            "<202403182259.64789700810.1@smtp-relay.mailin.fr>",
            "<202403182259.64789700810.2@smtp-relay.mailin.fr>",
        ]
    }

    def test_merge_data(self):
        self.set_mock_response(json_data=self._mock_batch_response)
        message = AnymailMessage(
            from_email="from@example.com",
            template_id=1234567,
            to=["alice@example.com", "Bob <bob@example.com>"],
            merge_data={
                "alice@example.com": {"name": "Alice", "group": "Developers"},
                "bob@example.com": {"name": "Bob"},  # and leave group undefined
                "nobody@example.com": {"name": "Not a recipient for this message"},
            },
            merge_global_data={"group": "Users", "site": "ExampleCo"},
        )
        message.send()

        # batch send uses same API endpoint as regular send:
        self.assert_esp_called("/v3/smtp/email")
        data = self.get_api_call_json()
        versions = data["messageVersions"]
        self.assertEqual(len(versions), 2)
        self.assertEqual(
            versions[0],
            {
                "to": [{"email": "alice@example.com"}],
                "params": {"name": "Alice", "group": "Developers"},
            },
        )
        self.assertEqual(
            versions[1],
            {
                "to": [{"email": "bob@example.com", "name": "Bob"}],
                "params": {"name": "Bob"},
            },
        )
        self.assertEqual(data["params"], {"group": "Users", "site": "ExampleCo"})

        recipients = message.anymail_status.recipients
        self.assertEqual(recipients["alice@example.com"].status, "queued")
        self.assertEqual(
            recipients["alice@example.com"].message_id,
            "<202403182259.64789700810.1@smtp-relay.mailin.fr>",
        )
        self.assertEqual(recipients["bob@example.com"].status, "queued")
        self.assertEqual(
            recipients["bob@example.com"].message_id,
            "<202403182259.64789700810.2@smtp-relay.mailin.fr>",
        )

    def test_merge_global_data(self):
        self.message.merge_global_data = {"a": "b"}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["params"], {"a": "b"})

    def test_merge_metadata(self):
        self.set_mock_response(json_data=self._mock_batch_response)
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123, "tier": "premium"},
            "bob@example.com": {"order_id": 678},
        }
        self.message.metadata = {"notification_batch": "zx912"}
        self.message.send()

        data = self.get_api_call_json()
        versions = data["messageVersions"]
        self.assertEqual(len(versions), 2)
        self.assertEqual(versions[0]["to"], [{"email": "alice@example.com"}])
        # metadata and merge_metadata[recipient] are combined:
        self.assertEqual(
            json.loads(versions[0]["headers"]["X-Mailin-custom"]),
            {"order_id": 123, "tier": "premium", "notification_batch": "zx912"},
        )
        self.assertEqual(
            versions[1]["to"], [{"name": "Bob", "email": "bob@example.com"}]
        )
        self.assertEqual(
            json.loads(versions[1]["headers"]["X-Mailin-custom"]),
            {"order_id": 678, "notification_batch": "zx912"},
        )
        # default metadata still sent in base headers:
        self.assertEqual(
            json.loads(data["headers"]["X-Mailin-custom"]),
            {"notification_batch": "zx912"},
        )

    def test_merge_headers(self):
        self.set_mock_response(json_data=self._mock_batch_response)
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.extra_headers = {
            "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
        }
        self.message.merge_headers = {
            "alice@example.com": {
                "List-Unsubscribe": "<https://example.com/a/>",
            },
            "bob@example.com": {
                "List-Unsubscribe": "<https://example.com/b/>",
            },
        }

        self.message.send()

        data = self.get_api_call_json()
        versions = data["messageVersions"]
        self.assertEqual(len(versions), 2)
        self.assertEqual(
            versions[0]["headers"], {"List-Unsubscribe": "<https://example.com/a/>"}
        )
        self.assertEqual(
            versions[1]["headers"], {"List-Unsubscribe": "<https://example.com/b/>"}
        )
        self.assertNotIn("params", versions[0])  # because no merge_data
        # non-merge headers still in base data
        self.assertEqual(
            data["headers"],
            {
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
        )

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("attachment", data)
        self.assertNotIn("bcc", data)
        self.assertNotIn("cc", data)
        self.assertNotIn("headers", data)
        self.assertNotIn("messageVersions", data)
        self.assertNotIn("params", data)
        self.assertNotIn("replyTo", data)
        self.assertNotIn("schedule", data)
        self.assertNotIn("tags", data)
        self.assertNotIn("templateId", data)

    def test_esp_extra(self):
        self.message.esp_extra = {
            "batchId": "5c6cfa04-eed9-42c2-8b5c-6d470d978e9d",
        }
        self.message.send()
        data = self.get_api_call_json()
        # merged from esp_extra:
        self.assertEqual(data["batchId"], "5c6cfa04-eed9-42c2-8b5c-6d470d978e9d")

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        # the DEFAULT_RAW_RESPONSE above is the *only* success response Brevo
        # returns, so no need to override it here
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.esp_response.content, self.DEFAULT_RAW_RESPONSE
        )

        self.assertEqual(
            msg.anymail_status.message_id,
            json.loads(msg.anymail_status.esp_response.content.decode("utf-8"))[
                "messageId"
            ],
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id,
            json.loads(msg.anymail_status.esp_response.content.decode("utf-8"))[
                "messageId"
            ],
        )

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.esp_extra = {"total": Decimal("19.99")}
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to Brevo", str(err))
        # original message
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")


@tag("brevo")
class BrevoBackendRecipientsRefusedTests(BrevoBackendMockAPITestCase):
    """
    Should raise AnymailRecipientsRefused when *all* recipients are rejected or invalid
    """

    # Brevo doesn't check email bounce or complaint lists at time of send --
    # it always just queues the message. You'll need to listen for the "rejected"
    # and "failed" events to detect refused recipients.
    pass  # not applicable to this backend


@tag("brevo")
class BrevoBackendSessionSharingTestCase(
    SessionSharingTestCases, BrevoBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("brevo")
@override_settings(EMAIL_BACKEND="anymail.backends.brevo.EmailBackend")
class BrevoBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test ESP backend without required settings in place"""

    def test_missing_auth(self):
        with self.assertRaisesRegex(AnymailConfigurationError, r"\bBREVO_API_KEY\b"):
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])



from ..exceptions import AnymailRequestsAPIError
from ..message import AnymailRecipientStatus
from ..utils import get_anymail_setting, update_deep
from .base_requests import AnymailRequestsBackend, RequestsPayload


class EmailBackend(AnymailRequestsBackend):
    """
    Mailjet API Email Backend
    """

    esp_name = "Mailjet"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        esp_name = self.esp_name
        self.api_key = get_anymail_setting(
            "api_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        self.secret_key = get_anymail_setting(
            "secret_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        api_url = get_anymail_setting(
            "api_url",
            esp_name=esp_name,
            kwargs=kwargs,
            default="https://api.mailjet.com/v3.1/",
        )
        if not api_url.endswith("/"):
            api_url += "/"
        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return MailjetPayload(message, defaults, self)

    def raise_for_status(self, response, payload, message):
        content_type = (
            response.headers.get("content-type", "").split(";")[0].strip().lower()
        )
        if 400 <= response.status_code <= 499 and content_type == "application/json":
            # Mailjet uses 4xx status codes for partial failure in batch send;
            # we'll determine how to handle below in parse_recipient_status.
            return
        super().raise_for_status(response, payload, message)

    def parse_recipient_status(self, response, payload, message):
        parsed_response = self.deserialize_json_response(response, payload, message)

        # Global error? (no messages sent)
        if "ErrorCode" in parsed_response:
            raise AnymailRequestsAPIError(
                email_message=message, payload=payload, response=response, backend=self
            )

        recipient_status = {}
        try:
            for result in parsed_response["Messages"]:
                # result["Status"] is "success" or "error"
                status = "sent" if result["Status"] == "success" else "failed"
                recipients = (
                    result.get("To", []) + result.get("Cc", []) + result.get("Bcc", [])
                )
                for recipient in recipients:
                    email = recipient["Email"]
                    # other Mailjet APIs expect MessageID (not MessageUUID)
                    message_id = str(recipient["MessageID"])
                    recipient_status[email] = AnymailRecipientStatus(
                        message_id=message_id, status=status
                    )
                # For errors, Mailjet doesn't identify the problem recipients. (This
                # can occur with a batch send.) Patch up the missing recipients below.
        except (KeyError, TypeError) as err:
            raise AnymailRequestsAPIError(
                "Invalid Mailjet API response format",
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            ) from err

        # Any recipient who wasn't reported as a 'success' must have been an error:
        for email in payload.recipients:
            if email.addr_spec not in recipient_status:
                recipient_status[email.addr_spec] = AnymailRecipientStatus(
                    message_id=None, status="failed"
                )

        return recipient_status


class MailjetPayload(RequestsPayload):
    def __init__(self, message, defaults, backend, *args, **kwargs):
        auth = (backend.api_key, backend.secret_key)
        http_headers = {
            "Content-Type": "application/json",
        }
        self.recipients = []  # for backend parse_recipient_status
        self.metadata = None
        super().__init__(
            message, defaults, backend, auth=auth, headers=http_headers, *args, **kwargs
        )

    def get_api_endpoint(self):
        return "send"

    def serialize_data(self):
        return self.serialize_json(self.data)

    #
    # Payload construction
    #

    def init_payload(self):
        # The v3.1 Send payload. We use Globals for most parameters,
        # which simplifies batch sending if it's used (and if not,
        # still works as expected for ordinary send).
        # https://dev.mailjet.com/email/reference/send-emails#v3_1_post_send
        self.data = {
            "Globals": {},
            "Messages": [],
        }

    def _burst_for_batch_send(self):
        """Expand the payload Messages into a separate object for each To address"""
        # This can be called multiple times -- if the payload has already been burst,
        # it will have no effect.
        # For simplicity, this assumes that "To" is the only Messages param we use
        # (because everything else goes in Globals).
        if len(self.data["Messages"]) == 1:
            to_recipients = self.data["Messages"][0].get("To", [])
            self.data["Messages"] = [{"To": [to]} for to in to_recipients]

    @staticmethod
    def _mailjet_email(email):
        """Expand an Anymail EmailAddress into Mailjet's {"Email", "Name"} dict"""
        result = {"Email": email.addr_spec}
        if email.display_name:
            result["Name"] = email.display_name
        return result

    def set_from_email(self, email):
        self.data["Globals"]["From"] = self._mailjet_email(email)

    def set_to(self, emails):
        # "To" is the one non-batch param we transmit in Messages rather than Globals.
        # (See also _burst_for_batch_send, set_merge_data, and set_merge_metadata.)
        if len(self.data["Messages"]) > 0:
            # This case shouldn't happen. Please file a bug report if it does.
            raise AssertionError("set_to called with non-empty Messages list")
        if emails:
            self.data["Messages"].append(
                {"To": [self._mailjet_email(email) for email in emails]}
            )
            self.recipients += emails
        else:
            # Mailjet requires a To list; cc-only messages aren't possible
            self.unsupported_feature("messages without any `to` recipients")

    def set_cc(self, emails):
        if emails:
            self.data["Globals"]["Cc"] = [
                self._mailjet_email(email) for email in emails
            ]
            self.recipients += emails

    def set_bcc(self, emails):
        if emails:
            self.data["Globals"]["Bcc"] = [
                self._mailjet_email(email) for email in emails
            ]
            self.recipients += emails

    def set_subject(self, subject):
        self.data["Globals"]["Subject"] = subject

    def set_reply_to(self, emails):
        if len(emails) > 0:
            self.data["Globals"]["ReplyTo"] = self._mailjet_email(emails[0])
            if len(emails) > 1:
                self.unsupported_feature("Multiple reply_to addresses")

    def set_extra_headers(self, headers):
        self.data["Globals"]["Headers"] = headers

    def set_text_body(self, body):
        if body:  # Django's default empty text body confuses Mailjet (esp. templates)
            self.data["Globals"]["TextPart"] = body

    def set_html_body(self, body):
        if body is not None:
            if "HTMLPart" in self.data["Globals"]:
                # second html body could show up through multiple alternatives,
                # or html body + alternative
                self.unsupported_feature("multiple html parts")

            self.data["Globals"]["HTMLPart"] = body

    def add_attachment(self, attachment):
        att = {
            "ContentType": attachment.mimetype,
            # Mailjet requires a non-empty Filename.
            "Filename": attachment.name or "attachment",
            "Base64Content": attachment.b64content,
        }
        if attachment.inline:
            field = "InlinedAttachments"
            att["ContentID"] = attachment.cid
        else:
            field = "Attachments"
        self.data["Globals"].setdefault(field, []).append(att)

    def set_envelope_sender(self, email):
        self.data["Globals"]["Sender"] = self._mailjet_email(email)

    def set_metadata(self, metadata):
        # Mailjet expects a single string payload
        self.data["Globals"]["EventPayload"] = self.serialize_json(metadata)
        self.metadata = metadata  # save for set_merge_metadata

    def set_merge_metadata(self, merge_metadata):
        self._burst_for_batch_send()
        for message in self.data["Messages"]:
            email = message["To"][0]["Email"]
            if email in merge_metadata:
                if self.metadata:
                    recipient_metadata = self.metadata.copy()
                    recipient_metadata.update(merge_metadata[email])
                else:
                    recipient_metadata = merge_metadata[email]
                message["EventPayload"] = self.serialize_json(recipient_metadata)

    def set_merge_headers(self, merge_headers):
        self._burst_for_batch_send()
        for message in self.data["Messages"]:
            email = message["To"][0]["Email"]
            if email in merge_headers:
                message["Headers"] = merge_headers[email]

    def set_tags(self, tags):
        # The choices here are CustomID or Campaign, and Campaign seems closer
        # to how "tags" are handled by other ESPs -- e.g., you can view dashboard
        # statistics across all messages with the same Campaign.
        if len(tags) > 0:
            self.data["Globals"]["CustomCampaign"] = tags[0]
            if len(tags) > 1:
                self.unsupported_feature("multiple tags (%r)" % tags)

    def set_track_clicks(self, track_clicks):
        self.data["Globals"]["TrackClicks"] = "enabled" if track_clicks else "disabled"

    def set_track_opens(self, track_opens):
        self.data["Globals"]["TrackOpens"] = "enabled" if track_opens else "disabled"

    def set_template_id(self, template_id):
        # Mailjet requires integer (not string) TemplateID:
        self.data["Globals"]["TemplateID"] = int(template_id)
        self.data["Globals"]["TemplateLanguage"] = True

    def set_merge_data(self, merge_data):
        self._burst_for_batch_send()
        for message in self.data["Messages"]:
            email = message["To"][0]["Email"]
            if email in merge_data:
                message["Variables"] = merge_data[email]

    def set_merge_global_data(self, merge_global_data):
        self.data["Globals"]["Variables"] = merge_global_data

    def set_esp_extra(self, extra):
        update_deep(self.data, extra)



from datetime import datetime, timezone
from email.mime.text import MIMEText

from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.core.mail import get_connection, send_mail
from django.test import SimpleTestCase
from django.test.utils import override_settings
from django.utils.functional import Promise
from django.utils.translation import gettext_lazy

from anymail.backends.test import EmailBackend as TestBackend, TestPayload
from anymail.exceptions import (
    AnymailConfigurationError,
    AnymailError,
    AnymailInvalidAddress,
    AnymailUnsupportedFeature,
)
from anymail.message import AnymailMessage
from anymail.utils import get_anymail_setting

from .utils import AnymailTestMixin


class SettingsTestBackend(TestBackend):
    """(useful only for these tests)"""

    def __init__(self, *args, **kwargs):
        esp_name = self.esp_name
        self.sample_setting = get_anymail_setting(
            "sample_setting", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        self.username = get_anymail_setting(
            "username", esp_name=esp_name, kwargs=kwargs, default=None, allow_bare=True
        )
        self.password = get_anymail_setting(
            "password", esp_name=esp_name, kwargs=kwargs, default=None, allow_bare=True
        )
        super().__init__(*args, **kwargs)


@override_settings(EMAIL_BACKEND="anymail.backends.test.EmailBackend")
class TestBackendTestCase(AnymailTestMixin, SimpleTestCase):
    """Base TestCase using Anymail's Test EmailBackend"""

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = AnymailMessage(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )

    @staticmethod
    def get_send_count():
        """Returns number of times "send api" has been called this test"""
        try:
            return len(mail.outbox)
        except AttributeError:
            # mail.outbox not initialized by either Anymail test
            # or Django locmem backend
            return 0

    @staticmethod
    def get_send_params():
        """Returns the params for the most recent "send api" call"""
        try:
            return mail.outbox[-1].anymail_test_params
        except IndexError:
            raise IndexError(
                "No messages have been sent through the Anymail test backend"
            )
        except AttributeError:
            raise AttributeError(
                "The last message sent was not processed"
                " through the Anymail test backend"
            )


@override_settings(EMAIL_BACKEND="tests.test_general_backend.SettingsTestBackend")
class BackendSettingsTests(TestBackendTestCase):
    """Test settings initializations for Anymail EmailBackends"""

    @override_settings(ANYMAIL={"TEST_SAMPLE_SETTING": "setting_from_anymail_settings"})
    def test_anymail_setting(self):
        """ESP settings usually come from ANYMAIL settings dict"""
        backend = get_connection()
        self.assertEqual(backend.sample_setting, "setting_from_anymail_settings")

    @override_settings(TEST_SAMPLE_SETTING="setting_from_bare_settings")
    def test_bare_setting(self):
        """ESP settings are also usually allowed at root of settings file"""
        backend = get_connection()
        self.assertEqual(backend.sample_setting, "setting_from_bare_settings")

    @override_settings(ANYMAIL={"TEST_SAMPLE_SETTING": "setting_from_settings"})
    def test_connection_kwargs_overrides_settings(self):
        """Can override settings file in get_connection"""
        backend = get_connection()
        self.assertEqual(backend.sample_setting, "setting_from_settings")

        backend = get_connection(sample_setting="setting_from_kwargs")
        self.assertEqual(backend.sample_setting, "setting_from_kwargs")

    def test_missing_setting(self):
        """Settings without defaults must be provided"""
        with self.assertRaises(AnymailConfigurationError) as cm:
            get_connection()
        self.assertIsInstance(cm.exception, ImproperlyConfigured)  # Django consistency
        errmsg = str(cm.exception)
        self.assertRegex(errmsg, r"\bTEST_SAMPLE_SETTING\b")
        self.assertRegex(errmsg, r"\bANYMAIL_TEST_SAMPLE_SETTING\b")

    @override_settings(
        ANYMAIL={
            "TEST_USERNAME": "username_from_settings",
            "TEST_PASSWORD": "password_from_settings",
            "TEST_SAMPLE_SETTING": "required",
        }
    )
    def test_username_password_kwargs_overrides(self):
        """Overrides for 'username' and 'password' should work like other overrides"""
        # These are special-cased because of default args in Django core mail functions.
        backend = get_connection()
        self.assertEqual(backend.username, "username_from_settings")
        self.assertEqual(backend.password, "password_from_settings")

        backend = get_connection(
            username="username_from_kwargs", password="password_from_kwargs"
        )
        self.assertEqual(backend.username, "username_from_kwargs")
        self.assertEqual(backend.password, "password_from_kwargs")


class UnsupportedFeatureTests(TestBackendTestCase):
    """Tests mail features not supported by backend are handled properly"""

    def test_unsupported_feature(self):
        """Unsupported features raise AnymailUnsupportedFeature"""
        # Test EmailBackend doesn't support non-HTML alternative parts
        self.message.attach_alternative(b"FAKE_MP3_DATA", "audio/mpeg")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    @override_settings(ANYMAIL={"IGNORE_UNSUPPORTED_FEATURES": True})
    def test_ignore_unsupported_features(self):
        """Setting prevents exception"""
        self.message.attach_alternative(b"FAKE_MP3_DATA", "audio/mpeg")
        self.message.send()  # should not raise exception


class SendDefaultsTests(TestBackendTestCase):
    """Tests backend support for global SEND_DEFAULTS and <ESP>_SEND_DEFAULTS"""

    @override_settings(
        ANYMAIL={
            "SEND_DEFAULTS": {
                # This isn't an exhaustive list of Anymail message attrs;
                # just one of each type
                "metadata": {"global": "globalvalue"},
                "send_at": datetime(2016, 5, 12, 4, 17, 0, tzinfo=timezone.utc),
                "tags": ["globaltag"],
                "template_id": "my-template",
                "track_clicks": True,
                "esp_extra": {"globalextra": "globalsetting"},
            }
        }
    )
    def test_send_defaults(self):
        """Test that (non-esp-specific) send defaults are applied"""
        self.message.send()
        params = self.get_send_params()
        # All these values came from ANYMAIL_SEND_DEFAULTS:
        self.assertEqual(params["metadata"], {"global": "globalvalue"})
        self.assertEqual(
            params["send_at"], datetime(2016, 5, 12, 4, 17, 0, tzinfo=timezone.utc)
        )
        self.assertEqual(params["tags"], ["globaltag"])
        self.assertEqual(params["template_id"], "my-template")
        self.assertEqual(params["track_clicks"], True)
        # Test EmailBackend merges esp_extra into params:
        self.assertEqual(params["globalextra"], "globalsetting")

    @override_settings(
        ANYMAIL={
            # SEND_DEFAULTS for the Test EmailBackend, because
            # "TEST" is the name of the Test EmailBackend's ESP
            "TEST_SEND_DEFAULTS": {
                "metadata": {"global": "espvalue"},
                "tags": ["esptag"],
                "track_opens": False,
                "esp_extra": {"globalextra": "espsetting"},
            }
        }
    )
    def test_esp_send_defaults(self):
        """Test that esp-specific send defaults are applied"""
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["metadata"], {"global": "espvalue"})
        self.assertEqual(params["tags"], ["esptag"])
        self.assertEqual(params["track_opens"], False)
        # Test EmailBackend merges esp_extra into params:
        self.assertEqual(params["globalextra"], "espsetting")

    @override_settings(
        ANYMAIL={
            "SEND_DEFAULTS": {
                "metadata": {"global": "globalvalue", "other": "othervalue"},
                "tags": ["globaltag"],
                "track_clicks": True,
                "track_opens": False,
                "esp_extra": {
                    "globalextra": "globalsetting",
                    "deepextra": {"deep1": "globaldeep1", "deep2": "globaldeep2"},
                },
            }
        }
    )
    def test_send_defaults_combine_with_message(self):
        """Individual message settings are *merged into* the global send defaults"""
        self.message.metadata = {"message": "messagevalue", "other": "override"}
        self.message.tags = ["messagetag"]
        self.message.track_clicks = False
        self.message.esp_extra = {
            "messageextra": "messagesetting",
            "deepextra": {"deep2": "messagedeep2", "deep3": "messagedeep3"},
        }

        self.message.send()
        params = self.get_send_params()
        self.assertEqual(
            params["metadata"],
            {  # metadata merged
                "global": "globalvalue",  # global default preserved
                "message": "messagevalue",  # message setting added
                "other": "override",  # message setting overrides global default
            },
        )
        # tags concatenated:
        self.assertEqual(params["tags"], ["globaltag", "messagetag"])
        self.assertEqual(params["track_clicks"], False)  # message overrides
        self.assertEqual(params["track_opens"], False)  # (no message setting)
        # esp_extra is deep merged:
        self.assertEqual(params["globalextra"], "globalsetting")
        self.assertEqual(params["messageextra"], "messagesetting")
        self.assertEqual(
            params["deepextra"],
            {"deep1": "globaldeep1", "deep2": "messagedeep2", "deep3": "messagedeep3"},
        )

        # Send another message to make sure original SEND_DEFAULTS unchanged
        send_mail("subject", "body", "from@example.com", ["to@example.com"])
        params = self.get_send_params()
        self.assertEqual(
            params["metadata"], {"global": "globalvalue", "other": "othervalue"}
        )
        self.assertEqual(params["tags"], ["globaltag"])
        self.assertEqual(params["track_clicks"], True)
        self.assertEqual(params["track_opens"], False)
        self.assertEqual(params["globalextra"], "globalsetting")
        self.assertEqual(
            params["deepextra"], {"deep1": "globaldeep1", "deep2": "globaldeep2"}
        )

    @override_settings(
        ANYMAIL={
            "SEND_DEFAULTS": {
                # This isn't an exhaustive list of Anymail message attrs;
                # just one of each type
                "metadata": {"global": "globalvalue"},
                "tags": ["globaltag"],
                "template_id": "global-template",
                "esp_extra": {"globalextra": "globalsetting"},
            },
            # "TEST" is the name of the Test EmailBackend's ESP
            "TEST_SEND_DEFAULTS": {
                "merge_global_data": {"esp": "espmerge"},
                "metadata": {"esp": "espvalue"},
                "tags": ["esptag"],
                "esp_extra": {"espextra": "espsetting"},
            },
        }
    )
    def test_esp_send_defaults_override_globals(self):
        """ESP-specific send defaults override *individual* global defaults"""
        self.message.send()
        params = self.get_send_params()
        # esp-defaults only:
        self.assertEqual(params["merge_global_data"], {"esp": "espmerge"})
        self.assertEqual(params["metadata"], {"esp": "espvalue"})
        self.assertEqual(params["tags"], ["esptag"])
        # global-defaults only:
        self.assertEqual(params["template_id"], "global-template")
        self.assertEqual(params["espextra"], "espsetting")
        # entire esp_extra is overriden by esp-send-defaults:
        self.assertNotIn("globalextra", params)


class LazyStringsTest(TestBackendTestCase):
    """
    Tests gettext_lazy strings forced real before passing to ESP transport.

    Docs notwithstanding, Django lazy strings *don't* work anywhere regular
    strings would. In particular, they aren't instances of unicode/str.
    There are some cases (e.g., urllib.urlencode, requests' _encode_params)
    where this can cause encoding errors or just very wrong results.

    Since Anymail sits on the border between Django app code and non-Django
    ESP code (e.g., requests), it's responsible for converting lazy text
    to actual strings.
    """

    def assertNotLazy(self, s, msg=None):
        self.assertNotIsInstance(s, Promise, msg=msg or "String %r is lazy" % str(s))

    def test_lazy_from(self):
        # This sometimes ends up lazy when settings.DEFAULT_FROM_EMAIL
        # is meant to be localized
        self.message.from_email = gettext_lazy('"Global Sales" <sales@example.com>')
        self.message.send()
        params = self.get_send_params()
        self.assertNotLazy(params["from"].address)

    def test_lazy_subject(self):
        self.message.subject = gettext_lazy("subject")
        self.message.send()
        params = self.get_send_params()
        self.assertNotLazy(params["subject"])

    def test_lazy_body(self):
        self.message.body = gettext_lazy("text body")
        self.message.attach_alternative(gettext_lazy("html body"), "text/html")
        self.message.send()
        params = self.get_send_params()
        self.assertNotLazy(params["text_body"])
        self.assertNotLazy(params["html_body"])

    def test_lazy_headers(self):
        self.message.extra_headers["X-Test"] = gettext_lazy("Test Header")
        self.message.send()
        params = self.get_send_params()
        self.assertNotLazy(params["extra_headers"]["X-Test"])

    def test_lazy_attachments(self):
        self.message.attach(
            gettext_lazy("test.csv"), gettext_lazy("test,csv,data"), "text/csv"
        )
        self.message.attach(MIMEText(gettext_lazy("contact info")))
        self.message.send()
        params = self.get_send_params()
        self.assertNotLazy(params["attachments"][0].name)
        self.assertNotLazy(params["attachments"][0].content)
        self.assertNotLazy(params["attachments"][1].content)

    def test_lazy_tags(self):
        self.message.tags = [gettext_lazy("Shipping"), gettext_lazy("Sales")]
        self.message.send()
        params = self.get_send_params()
        self.assertNotLazy(params["tags"][0])
        self.assertNotLazy(params["tags"][1])

    def test_lazy_metadata(self):
        self.message.metadata = {"order_type": gettext_lazy("Subscription")}
        self.message.send()
        params = self.get_send_params()
        self.assertNotLazy(params["metadata"]["order_type"])

    def test_lazy_merge_data(self):
        self.message.merge_data = {
            "to@example.com": {"duration": gettext_lazy("One Month")}
        }
        self.message.merge_global_data = {"order_type": gettext_lazy("Subscription")}
        self.message.send()
        params = self.get_send_params()
        self.assertNotLazy(params["merge_data"]["to@example.com"]["duration"])
        self.assertNotLazy(params["merge_global_data"]["order_type"])


class CatchCommonErrorsTests(TestBackendTestCase):
    """Anymail should catch and provide useful errors for common mistakes"""

    def test_explains_reply_to_must_be_list(self):
        """reply_to must be a list (or other iterable), not a single string"""
        # Django's EmailMessage.__init__ catches this and warns, but isn't
        # involved if you assign attributes later. Anymail should catch that case.
        # (This also applies to to, cc, and bcc, but Django stumbles over those cases
        # in EmailMessage.recipients (called from EmailMessage.send) before
        # Anymail gets a chance to complain.)
        self.message.reply_to = "single-reply-to@example.com"
        with self.assertRaisesMessage(
            TypeError, '"reply_to" attribute must be a list or other iterable'
        ):
            self.message.send()

    def test_explains_reply_to_must_be_list_lazy(self):
        """Same as previous tests, with lazy strings"""
        # Lazy strings can fool string/iterable detection
        self.message.reply_to = gettext_lazy("single-reply-to@example.com")
        with self.assertRaisesMessage(
            TypeError, '"reply_to" attribute must be a list or other iterable'
        ):
            self.message.send()

    def test_identifies_source_of_parsing_errors(self):
        """Errors parsing email addresses should say which field had the problem"""
        # Note: General email address parsing tests are in
        # test_utils.ParseAddressListTests. This just checks the error includes the
        # field name when parsing for sending a message.
        self.message.from_email = ""
        with self.assertRaisesMessage(
            AnymailInvalidAddress,
            "Invalid email address '' parsed from '' in `from_email`.",
        ):
            self.message.send()
        self.message.from_email = "from@example.com"

        # parse_address_list
        self.message.to = ["ok@example.com", "oops"]
        with self.assertRaisesMessage(
            AnymailInvalidAddress,
            "Invalid email address 'oops' parsed from 'ok@example.com, oops' in `to`.",
        ):
            self.message.send()
        self.message.to = ["test@example.com"]

        # parse_single_address
        self.message.envelope_sender = "one@example.com, two@example.com"
        with self.assertRaisesMessage(
            AnymailInvalidAddress,
            "Only one email address is allowed; found 2"
            " in 'one@example.com, two@example.com' in `envelope_sender`.",
        ):
            self.message.send()
        delattr(self.message, "envelope_sender")

        # process_extra_headers
        self.message.extra_headers["From"] = "Mail, Inc. <mail@example.com>"
        with self.assertRaisesMessage(
            AnymailInvalidAddress,
            "Invalid email address 'Mail' parsed from"
            " 'Mail, Inc. <mail@example.com>' in `extra_headers['From']`."
            " (Maybe missing quotes around a display-name?)",
        ):
            self.message.send()

    def test_error_minimizes_pii_leakage(self):
        """
        AnymailError messages should generally avoid including
        email addresses where not relevant to the error.

        (This is not a guarantee that exceptions will never include
        email addresses or other PII. The ESP's own error--which *is*
        deliberately included in the message--will often include the
        email address, and Anymail makes no attempt to filter that.)
        """
        # Cause an error (not related to the specific email addresses involved):
        self.message.attach_alternative("...", "audio/mpeg4")
        with self.assertRaises(AnymailError) as cm:
            self.message.send()
        error = cm.exception
        self.assertNotIn("from@example.com", str(error))
        self.assertNotIn("to@example.com", str(error))


def flatten_emails(emails):
    return [str(email) for email in emails]


class SpecialHeaderTests(TestBackendTestCase):
    """Anymail should handle special extra_headers the same way Django does"""

    def test_reply_to(self):
        """
        Django allows message.reply_to and message.extra_headers['Reply-To'],
        and the latter takes precedence
        """
        self.message.reply_to = ["attr@example.com"]
        self.message.extra_headers = {"X-Extra": "extra"}
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(flatten_emails(params["reply_to"]), ["attr@example.com"])
        self.assertEqual(params["extra_headers"], {"X-Extra": "extra"})

        self.message.reply_to = None
        self.message.extra_headers = {
            "Reply-To": "header@example.com",
            "X-Extra": "extra",
        }
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(flatten_emails(params["reply_to"]), ["header@example.com"])
        # Reply-To no longer there:
        self.assertEqual(params["extra_headers"], {"X-Extra": "extra"})

        # If both are supplied, the header wins (to match Django EmailMessage.message()
        # behavior). Also, header names are case-insensitive.
        self.message.reply_to = ["attr@example.com"]
        self.message.extra_headers = {
            "REPLY-to": "header@example.com",
            "X-Extra": "extra",
        }
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(flatten_emails(params["reply_to"]), ["header@example.com"])
        # Reply-To no longer there
        self.assertEqual(params["extra_headers"], {"X-Extra": "extra"})

    def test_envelope_sender(self):
        """
        Django treats message.from_email as envelope-sender
        if message.extra_headers['From'] is set
        """
        # Using Anymail's envelope_sender extension
        self.message.from_email = "Header From <header@example.com>"
        self.message.envelope_sender = (
            "Envelope From <envelope@bounces.example.com>"  # Anymail extension
        )
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["from"].address, "Header From <header@example.com>")
        self.assertEqual(params["envelope_sender"], "envelope@bounces.example.com")

        # Using Django's undocumented message.extra_headers['From'] extension
        # (see https://code.djangoproject.com/ticket/9214)
        self.message.from_email = "Envelope From <envelope@bounces.example.com>"
        self.message.extra_headers = {"From": "Header From <header@example.com>"}
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["from"].address, "Header From <header@example.com>")
        self.assertEqual(params["envelope_sender"], "envelope@bounces.example.com")
        # From was removed from extra-headers:
        self.assertNotIn("From", params.get("extra_headers", {}))

    def test_spoofed_to_header(self):
        """
        Django treats message.to as envelope-recipient
        if message.extra_headers['To'] is set
        """
        # No current ESP supports this (and it's unlikely they would)
        self.message.to = ["actual-recipient@example.com"]
        self.message.extra_headers = {
            "To": "Apparent Recipient <but-not-really@example.com>"
        }
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "spoofing `To` header"
        ):
            self.message.send()


class AlternativePartsTests(TestBackendTestCase):
    """
    Anymail should handle alternative parts consistently with Django's SMTP backend
    """

    def test_default_usage(self):
        """Body defaults to text/plain, use alternative for html"""
        self.message.body = "plain body"
        self.message.attach_alternative("html body", "text/html")
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["text_body"], "plain body")
        self.assertEqual(params["html_body"], "html body")
        self.assertNotIn("alternatives", params)

    def test_content_subtype_html(self):
        """Change body to text/html, use alternative for plain"""
        self.message.content_subtype = "html"
        self.message.body = "html body"
        self.message.attach_alternative("plain body", "text/plain")
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["text_body"], "plain body")
        self.assertEqual(params["html_body"], "html body")
        self.assertNotIn("alternatives", params)

    def test_attach_plain_and_html(self):
        """Use alternatives for both bodies"""
        message = AnymailMessage(
            subject="Subject", from_email="from@example.com", to=["to@example.com"]
        )
        message.attach_alternative("plain body", "text/plain")
        message.attach_alternative("html body", "text/html")
        message.send()
        params = self.get_send_params()
        self.assertEqual(params["text_body"], "plain body")
        self.assertEqual(params["html_body"], "html body")
        self.assertNotIn("alternatives", params)

    def test_additional_plain_part(self):
        """Two plaintext bodies"""
        # In theory this is supported (e.g., for different languages or charsets),
        # though MUAs are unlikely to display anything after the first.
        self.message.body = "plain body"
        self.message.attach_alternative("second plain body", "text/plain")
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["text_body"], "plain body")
        self.assertEqual(params["alternatives"], [("second plain body", "text/plain")])

    def test_exotic_content_subtype(self):
        """Change body to text/calendar, use alternatives for plain and html"""
        # This is unlikely to work with most ESPs, but we can try to communicate the
        # intent... (You probably want an attachment rather than an alternative part.)
        self.message.content_subtype = "calendar"
        self.message.body = "BEGIN:VCALENDAR..."
        self.message.attach_alternative("plain body", "text/plain")
        self.message.attach_alternative("html body", "text/html")
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["text_body"], "plain body")
        self.assertEqual(params["html_body"], "html body")
        self.assertEqual(
            params["alternatives"], [("BEGIN:VCALENDAR...", "text/calendar")]
        )


class BatchSendDetectionTestCase(TestBackendTestCase):
    """Tests shared code to consistently determine whether to use batch send"""

    def test_default_is_not_batch(self):
        self.message.send()
        params = self.get_send_params()
        self.assertFalse(params["is_batch_send"])

    def test_merge_data_implies_batch(self):
        self.message.merge_data = {}  # *anything* (even empty dict) implies batch
        self.message.send()
        params = self.get_send_params()
        self.assertTrue(params["is_batch_send"])

    def test_merge_metadata_implies_batch(self):
        self.message.merge_metadata = {}  # *anything* (even empty dict) implies batch
        self.message.send()
        params = self.get_send_params()
        self.assertTrue(params["is_batch_send"])

    def test_merge_global_data_does_not_imply_batch(self):
        self.message.merge_global_data = {}
        self.message.send()
        params = self.get_send_params()
        self.assertFalse(params["is_batch_send"])

    def test_cannot_call_is_batch_during_init(self):
        # It's tempting to try to warn about unsupported batch features in setters,
        # but because of the way payload attrs are processed, it won't work...
        class ImproperlyImplementedPayload(TestPayload):
            def set_cc(self, emails):
                if self.is_batch():  # this won't work here!
                    self.unsupported_feature("cc with batch send")
                super().set_cc(emails)

        connection = mail.get_connection(
            "anymail.backends.test.EmailBackend",
            payload_class=ImproperlyImplementedPayload,
        )
        with self.assertRaisesMessage(
            AssertionError, "Cannot call is_batch before all attributes processed"
        ):
            connection.send_messages([self.message])



from datetime import datetime

from ..exceptions import AnymailRequestsAPIError
from ..message import ANYMAIL_STATUSES, AnymailRecipientStatus
from ..utils import get_anymail_setting
from .base_requests import AnymailRequestsBackend, RequestsPayload


class EmailBackend(AnymailRequestsBackend):
    """
    Mandrill API Email Backend
    """

    esp_name = "Mandrill"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        esp_name = self.esp_name
        self.api_key = get_anymail_setting(
            "api_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        api_url = get_anymail_setting(
            "api_url",
            esp_name=esp_name,
            kwargs=kwargs,
            default="https://mandrillapp.com/api/1.0",
        )
        if not api_url.endswith("/"):
            api_url += "/"
        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return MandrillPayload(message, defaults, self)

    def parse_recipient_status(self, response, payload, message):
        parsed_response = self.deserialize_json_response(response, payload, message)
        recipient_status = {}
        try:
            # Mandrill returns a list of { email, status, _id, reject_reason }
            # for each recipient
            for item in parsed_response:
                email = item["email"]
                status = item["status"]
                if status not in ANYMAIL_STATUSES:
                    status = "unknown"
                # "_id" can be missing for invalid/rejected recipients:
                message_id = item.get("_id", None)
                recipient_status[email] = AnymailRecipientStatus(
                    message_id=message_id, status=status
                )
        except (KeyError, TypeError) as err:
            raise AnymailRequestsAPIError(
                "Invalid Mandrill API response format",
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            ) from err
        return recipient_status


def encode_date_for_mandrill(dt):
    """Format a datetime for use as a Mandrill API date field

    Mandrill expects "YYYY-MM-DD HH:MM:SS" in UTC
    """
    if isinstance(dt, datetime):
        dt = dt.replace(microsecond=0)
        if dt.utcoffset() is not None:
            dt = (dt - dt.utcoffset()).replace(tzinfo=None)
        return dt.isoformat(" ")
    else:
        return dt


class MandrillPayload(RequestsPayload):
    def __init__(self, *args, **kwargs):
        self.esp_extra = {}  # late-bound in serialize_data
        super().__init__(*args, **kwargs)

    def get_api_endpoint(self):
        if "template_name" in self.data:
            return "messages/send-template.json"
        else:
            return "messages/send.json"

    def serialize_data(self):
        self.process_esp_extra()
        if self.is_batch():
            # hide recipients from each other
            self.data["message"]["preserve_recipients"] = False
        return self.serialize_json(self.data)

    #
    # Payload construction
    #

    def init_payload(self):
        self.data = {
            "key": self.backend.api_key,
            "message": {},
        }

    def set_from_email(self, email):
        self.data["message"]["from_email"] = email.addr_spec
        if email.display_name:
            self.data["message"]["from_name"] = email.display_name

    def add_recipient(self, recipient_type, email):
        assert recipient_type in ["to", "cc", "bcc"]
        recipient_data = {"email": email.addr_spec, "type": recipient_type}
        if email.display_name:
            recipient_data["name"] = email.display_name
        to_list = self.data["message"].setdefault("to", [])
        to_list.append(recipient_data)

    def set_subject(self, subject):
        self.data["message"]["subject"] = subject

    def set_reply_to(self, emails):
        if emails:
            reply_to = ", ".join([str(email) for email in emails])
            self.data["message"].setdefault("headers", {})["Reply-To"] = reply_to

    def set_extra_headers(self, headers):
        self.data["message"].setdefault("headers", {}).update(headers)

    def set_text_body(self, body):
        self.data["message"]["text"] = body

    def set_html_body(self, body):
        if "html" in self.data["message"]:
            # second html body could show up through multiple alternatives,
            # or html body + alternative
            self.unsupported_feature("multiple html parts")
        self.data["message"]["html"] = body

    def add_attachment(self, attachment):
        if attachment.inline:
            field = "images"
            name = attachment.cid
        else:
            field = "attachments"
            name = attachment.name or ""
        self.data["message"].setdefault(field, []).append(
            {
                "type": attachment.mimetype,
                "name": name,
                "content": attachment.b64content,
            }
        )

    def set_envelope_sender(self, email):
        # Only the domain is used
        self.data["message"]["return_path_domain"] = email.domain

    def set_metadata(self, metadata):
        self.data["message"]["metadata"] = metadata

    def set_send_at(self, send_at):
        self.data["send_at"] = encode_date_for_mandrill(send_at)

    def set_tags(self, tags):
        self.data["message"]["tags"] = tags

    def set_track_clicks(self, track_clicks):
        self.data["message"]["track_clicks"] = track_clicks

    def set_track_opens(self, track_opens):
        self.data["message"]["track_opens"] = track_opens

    def set_template_id(self, template_id):
        self.data["template_name"] = template_id
        self.data.setdefault("template_content", [])  # Mandrill requires something here

    def set_merge_data(self, merge_data):
        self.data["message"]["merge_vars"] = [
            {
                "rcpt": rcpt,
                "vars": [
                    # sort for testing reproducibility:
                    {"name": key, "content": rcpt_data[key]}
                    for key in sorted(rcpt_data.keys())
                ],
            }
            for rcpt, rcpt_data in merge_data.items()
        ]

    def set_merge_global_data(self, merge_global_data):
        self.data["message"]["global_merge_vars"] = [
            {"name": var, "content": value} for var, value in merge_global_data.items()
        ]

    def set_merge_metadata(self, merge_metadata):
        # recipient_metadata format is similar to, but not quite the same as,
        # merge_vars:
        self.data["message"]["recipient_metadata"] = [
            {"rcpt": rcpt, "values": rcpt_data}
            for rcpt, rcpt_data in merge_metadata.items()
        ]

    def set_esp_extra(self, extra):
        # late bind in serialize_data, so that obsolete Djrill attrs can contribute
        self.esp_extra = extra

    def process_esp_extra(self):
        if self.esp_extra is not None and len(self.esp_extra) > 0:
            esp_extra = self.esp_extra
            # Convert pythonic template_content dict to Mandrill name/content list
            try:
                template_content = esp_extra["template_content"]
            except KeyError:
                pass
            else:
                # if it's dict-like:
                if hasattr(template_content, "items"):
                    if esp_extra is self.esp_extra:
                        # don't modify caller's value
                        esp_extra = self.esp_extra.copy()
                    esp_extra["template_content"] = [
                        {"name": var, "content": value}
                        for var, value in template_content.items()
                    ]
            # Convert pythonic recipient_metadata dict to Mandrill rcpt/values list
            try:
                recipient_metadata = esp_extra["message"]["recipient_metadata"]
            except KeyError:
                pass
            else:
                # if it's dict-like:
                if hasattr(recipient_metadata, "keys"):
                    if esp_extra["message"] is self.esp_extra["message"]:
                        # don't modify caller's value:
                        esp_extra["message"] = self.esp_extra["message"].copy()
                    # For testing reproducibility, sort the recipients
                    esp_extra["message"]["recipient_metadata"] = [
                        {"rcpt": rcpt, "values": recipient_metadata[rcpt]}
                        for rcpt in sorted(recipient_metadata.keys())
                    ]
            # Merge esp_extra with payload data: shallow merge within ['message']
            # and top-level keys
            self.data.update({k: v for k, v in esp_extra.items() if k != "message"})
            try:
                self.data["message"].update(esp_extra["message"])
            except KeyError:
                pass



import os
import unittest
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin, sample_image_path

# For most integration tests, Postmark's sandboxed "POSTMARK_API_TEST" token is used.
# But to test template sends, a real Postmark server token and template id are needed:
ANYMAIL_TEST_POSTMARK_SERVER_TOKEN = os.getenv("ANYMAIL_TEST_POSTMARK_SERVER_TOKEN")
ANYMAIL_TEST_POSTMARK_TEMPLATE_ID = os.getenv("ANYMAIL_TEST_POSTMARK_TEMPLATE_ID")
ANYMAIL_TEST_POSTMARK_DOMAIN = os.getenv("ANYMAIL_TEST_POSTMARK_DOMAIN")


@tag("postmark", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_POSTMARK_DOMAIN,
    "Set ANYMAIL_TEST_POSTMARK_DOMAIN environment variable "
    "to run Postmark template integration tests",
)
@override_settings(
    ANYMAIL_POSTMARK_SERVER_TOKEN="POSTMARK_API_TEST",
    EMAIL_BACKEND="anymail.backends.postmark.EmailBackend",
)
class PostmarkBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """Postmark API integration tests

    These tests run against the **live** Postmark API, but using a
    test key that's not capable of sending actual email.
    """

    def setUp(self):
        super().setUp()
        self.from_email = "from@%s" % ANYMAIL_TEST_POSTMARK_DOMAIN
        self.message = AnymailMessage(
            "Anymail Postmark integration test",
            "Text content",
            self.from_email,
            ["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the Postmark send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["test+to1@anymail.dev"].status
        message_id = anymail_status.recipients["test+to1@anymail.dev"].message_id

        self.assertEqual(sent_status, "sent")
        self.assertGreater(len(message_id), 0)  # non-empty string
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        message = AnymailMessage(
            subject="Anymail Postmark all-options integration test",
            body="This is the text body",
            from_email=formataddr(("Test From, with comma", self.from_email)),
            to=["test+to1@anymail.dev", "Recipient 2 <test+to2@anymail.dev>"],
            cc=["test+cc1@anymail.dev", "Copy 2 <test+cc2@anymail.dev>"],
            bcc=["test+bcc1@anymail.dev", "Blind Copy 2 <test+bcc2@anymail.dev>"],
            reply_to=["reply1@example.com", "Reply 2 <reply2@example.com>"],
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
            # no send_at support
            metadata={"meta1": "simple string", "meta2": 2},
            tags=["tag 1"],  # max one tag
            track_opens=True,
            track_clicks=True,
            # either of these merge_ options will force batch send
            # (unique message for each "to" recipient)
            merge_metadata={
                "test+to1@anymail.dev": {"customer-id": "ZXK9123"},
                "test+to2@anymail.dev": {"customer-id": "ZZT4192"},
            },
            merge_headers={
                "test+to1@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                },
                "test+to2@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")
        cid = message.attach_inline_image_file(sample_image_path())
        message.attach_alternative(
            "<p><b>HTML:</b> with <a href='http://example.com'>link</a>"
            "and image: <img src='cid:%s'></div>" % cid,
            "text/html",
        )

        message.send()
        self.assertEqual(message.anymail_status.status, {"sent"})
        self.assertEqual(
            message.anymail_status.recipients["test+to1@anymail.dev"].status, "sent"
        )
        self.assertEqual(
            message.anymail_status.recipients["test+to2@anymail.dev"].status, "sent"
        )
        # distinct messages should have different message_ids:
        self.assertNotEqual(
            message.anymail_status.recipients["test+to1@anymail.dev"].message_id,
            message.anymail_status.recipients["test+to2@anymail.dev"].message_id,
        )

    def test_invalid_from(self):
        self.message.from_email = "webmaster@localhost"  # Django's default From
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertEqual(err.status_code, 422)
        self.assertIn("Invalid 'From' address", str(err))

    @unittest.skipUnless(
        ANYMAIL_TEST_POSTMARK_SERVER_TOKEN
        and ANYMAIL_TEST_POSTMARK_TEMPLATE_ID
        and ANYMAIL_TEST_POSTMARK_DOMAIN,
        "Set ANYMAIL_TEST_POSTMARK_SERVER_TOKEN and ANYMAIL_TEST_POSTMARK_TEMPLATE_ID "
        "and ANYMAIL_TEST_POSTMARK_DOMAIN environment variables to run Postmark "
        "template integration tests",
    )
    @override_settings(ANYMAIL_POSTMARK_SERVER_TOKEN=ANYMAIL_TEST_POSTMARK_SERVER_TOKEN)
    def test_template(self):
        message = AnymailMessage(
            from_email=self.from_email,
            to=["test+to1@anymail.dev", "Second Recipient <test+to2@anymail.dev>"],
            template_id=ANYMAIL_TEST_POSTMARK_TEMPLATE_ID,
            merge_data={
                "test+to1@anymail.dev": {"name": "Recipient 1", "order_no": "12345"},
                "test+to2@anymail.dev": {"order_no": "6789"},
            },
            merge_global_data={"name": "Valued Customer"},
        )
        message.send()
        self.assertEqual(message.anymail_status.status, {"sent"})

    @override_settings(ANYMAIL_POSTMARK_SERVER_TOKEN="Hey, that's not a server token!")
    def test_invalid_server_token(self):
        # Message will include something like
        # "Request does not contain a valid Server token"
        # or "Please verify that you are using a valid token"
        with self.assertRaisesRegex(AnymailAPIError, r"valid.*token"):
            self.message.send()
